from ._metadata import __version__
from .config import _BACKEND, BACKEND_PRETTY_NAME, PYTORCH_BACKEND, TENSORFLOW_BACKEND

BACKEND_IMPORT_ERROR_MESSAGE = f"""
{BACKEND_PRETTY_NAME} is selected as backend, but it is not installed.
Please install DLOmix with {BACKEND_PRETTY_NAME} using 'pip install dlomix[{BACKEND_PRETTY_NAME.lower()}]'
or separately install the required backend via pip.
"""


if _BACKEND in PYTORCH_BACKEND:
    try:
        import torch  # Check for PyTorch
    except ImportError:
        raise ImportError(BACKEND_IMPORT_ERROR_MESSAGE)
if _BACKEND in TENSORFLOW_BACKEND:
    try:
        import tensorflow  # Check for TensorFlow
    except ImportError:
        raise ImportError(BACKEND_IMPORT_ERROR_MESSAGE)

    from ._register_keras import *  # Register Keras custom objects

__all__ = [
    "_BACKEND",  # Also expose the current backend name
    "__version__",  # Expose the version of the package
]



import logging

import torch

from dlomix.models.chargestate import ChargeStatePredictor
from dlomix.models.chargestate_torch import (
    ChargeStatePredictor as ChargeStatePredictorTorch,
)
from dlomix.models.prosit import PrositIntensityPredictor, PrositRetentionTimePredictor
from dlomix.models.prosit_torch import (
    PrositIntensityPredictor as PrositIntensityPredictorTorch,
)
from dlomix.models.prosit_torch import (
    PrositRetentionTimePredictor as PrositRetentionTimePredictorTorch,
)

logger = logging.getLogger(__name__)


def basic_model_existence_test_torch(model):
    logger.info(model)
    assert model is not None

    assert len(list(model.parameters())) > 0


# ------------------ CS | check for existence of model & its parameters ------------------


def test_dominant_chargestate_model_torch():
    model = ChargeStatePredictorTorch(model_flavour="dominant")
    basic_model_existence_test_torch(model)


def test_observed_chargestate_model_torch():
    model = ChargeStatePredictorTorch(model_flavour="observed")
    basic_model_existence_test_torch(model)


def test_chargestate_distribution_model_torch():
    model = ChargeStatePredictorTorch(model_flavour="relative")
    basic_model_existence_test_torch(model)


# ------------------ CS | comparison of tf & torch ------------------


def test_tf_torch_equivalence_chargestate_model_shapes():
    # to compare tf & torch: input & output shapes at beginnin & end of 1 forward

    batch_size = 2
    seq_len = 30

    dummy_input_torch = torch.randint(low=0, high=15, size=(batch_size, seq_len))
    dummi_input_tf = dummy_input_torch.numpy()

    model_tf = ChargeStatePredictor(model_flavour="dominant", seq_length=seq_len)
    model_torch = ChargeStatePredictorTorch(
        model_flavour="dominant", seq_length=seq_len
    )

    output_tf = model_tf(dummi_input_tf)
    output_torch = model_torch(dummy_input_torch)

    assert output_tf.shape == output_torch.detach().numpy().shape


# -------------- Prosit RT | check for existence of model & its parameters -------


def test_RT_model_torch():
    model = PrositRetentionTimePredictorTorch()
    basic_model_existence_test_torch(model)


# -------------- Prosit RT | comparison of tf & torch ----------------------


def test_tf_torch_equivalence_RT_model_shapes():
    # to compare tf & torch: input & output shapes at beginnin & end of 1 forward

    batch_size = 2
    seq_len = 30

    dummy_input_torch = torch.randint(low=0, high=15, size=(batch_size, seq_len))
    dummi_input_tf = dummy_input_torch.numpy()

    model_tf = PrositRetentionTimePredictor(seq_length=seq_len)
    model_torch = PrositRetentionTimePredictorTorch(seq_length=seq_len)

    output_tf = model_tf(dummi_input_tf)
    output_torch = model_torch(dummy_input_torch)

    assert output_tf.shape == output_torch.detach().numpy().shape


# -------------- Prosit Intensity | check for existence of model & its parameters -------
def test_intensity_model_torch():
    model = PrositIntensityPredictorTorch()
    basic_model_existence_test_torch(model)


# -------------- Prosit Intensity | comparison of tf & torch ----------------------
def test_tf_torch_equivalence_intensity_model_shapes():
    # to compare tf & torch: input & output shapes at beginnin & end of 1 forward

    batch_size = 2
    seq_len = 30

    dummy_input_torch = torch.randint(low=0, high=15, size=(batch_size, seq_len))
    dummi_input_tf = dummy_input_torch.numpy()

    model_tf = PrositIntensityPredictor(
        seq_length=seq_len,
    )

    model_torch = PrositIntensityPredictorTorch(
        seq_length=seq_len,
    )

    output_tf = model_tf(dummi_input_tf)
    output_torch = model_torch(dummy_input_torch)

    assert output_tf.shape == output_torch.detach().numpy().shape



import numpy as np

DEFAULT_PARQUET_ENGINE = "pyarrow"

retention_time_pipeline_parameters = {
    "model_params": {"seq_length": 30},
    "data_params": {
        "seq_length": 30,
    },
    "trained_model_path": "../pretrained_models/retention_time/example_rtmodel/",
    "trained_model_zipfile_name": "rtmodel.zip",
    "trained_model_stats": [0.0, 1.0],
}

retention_time_pipeline_parameters.update(
    {
        "trained_model_url": "https://raw.githubusercontent.com/wilhelm-lab/dlomix/develop"
        + retention_time_pipeline_parameters["trained_model_path"].strip("..")
        + retention_time_pipeline_parameters["trained_model_zipfile_name"]
    }
)

_ALPHABET_UNMOD_ORDERED = "ACDEFGHIKLMNPQRSTVWY"
ALPHABET_UNMOD = {k: v for v, k in enumerate(_ALPHABET_UNMOD_ORDERED, start=1)}
ALPHABET_UNMOD.update({"[]-": 21, "-[]": 22})
ALPHABET_UNMOD.update({"[UNIMOD:737]-": 21})

_ALPHABET_NAIVE_MODS_ORDERED = ["C[UNIMOD:4]", "M[UNIMOD:35]"]
ALPHABET_NAIVE_MODS = {
    k: v
    for v, k in enumerate(_ALPHABET_NAIVE_MODS_ORDERED, start=len(ALPHABET_UNMOD) + 1)
}
ALPHABET_NAIVE_MODS.update(ALPHABET_UNMOD)

PTMS_ALPHABET = {
    "A": 1,
    "C": 2,
    "D": 3,
    "E": 4,
    "F": 5,
    "G": 6,
    "H": 7,
    "I": 8,
    "K": 9,
    "L": 10,
    "M": 11,
    "N": 12,
    "P": 13,
    "Q": 14,
    "R": 15,
    "S": 16,
    "T": 17,
    "V": 18,
    "W": 19,
    "Y": 20,
    "[]-": 21,
    "-[]": 22,
    "[UNIMOD:737]-": 56,
    "M[UNIMOD:35]": 23,
    "S[UNIMOD:21]": 24,
    "T[UNIMOD:21]": 25,
    "Y[UNIMOD:21]": 26,
    "R[UNIMOD:7]": 27,
    "Q[UNIMOD:7]": 4,
    "N[UNIMOD:7]": 3,
    "K[UNIMOD:1]": 28,
    "K[UNIMOD:121]": 29,
    "Q[UNIMOD:28]": 30,
    "R[UNIMOD:34]": 31,
    "K[UNIMOD:34]": 32,
    "T[UNIMOD:43]": 35,
    "S[UNIMOD:43]": 36,
    "C[UNIMOD:4]": 37,
    "[UNIMOD:1]-": 38,
    "E[UNIMOD:27]": 39,
    "K[UNIMOD:36]": 40,
    "K[UNIMOD:37]": 41,
    "K[UNIMOD:122]": 42,
    "K[UNIMOD:58]": 43,
    "K[UNIMOD:1289]": 44,
    "K[UNIMOD:747]": 45,
    "K[UNIMOD:64]": 46,
    "K[UNIMOD:1848]": 47,
    "K[UNIMOD:1363]": 48,
    "K[UNIMOD:1849]": 49,
    "K[UNIMOD:3]": 50,
    "K[UNIMOD:737]": 55,
    "R[UNIMOD:36]": 51,
    "R[UNIMOD:36a]": 52,
    "P[UNIMOD:35]": 53,
    "Y[UNIMOD:354]": 54,
}

# ToDo: double check chargestate data additional tokens for the alphabet

# PTMS_ALPHABET['W[UNIMOD:425]'] = 57
# PTMS_ALPHABET['K[UNIMOD:1342]'] = 58
# PTMS_ALPHABET['[UNIMOD:27]-'] = 59


# ---- detectability_model_constants.py ----
CLASSES_LABELS = ["Non-Flyer", "Weak Flyer", "Intermediate Flyer", "Strong Flyer"]

alphabet = [
    "0",
    "A",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "V",
    "W",
    "Y",
]

aa_to_int_dict = dict((aa, i) for i, aa in enumerate(alphabet))

int_to_aa_dict = dict((i, aa) for i, aa in enumerate(alphabet))


padding_char = np.zeros(len(alphabet))
padding_char[0] = 1
# ------- detectability_model_constants ------



from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray

from .dataset import PeptideDataset
from .dataset_config import DatasetConfig
from .dataset_utils import EncodingScheme


class IonMobilityDataset(PeptideDataset):
    """
    A dataset class for handling ion mobility data.

    Args:
        data_source (Optional[Union[str, List]]): The data source for the dataset. Defaults to None.
        val_data_source (Optional[Union[str, List]]): The validation data source for the dataset. Defaults to None.
        test_data_source (Optional[Union[str, List]]): The test data source for the dataset. Defaults to None.
        data_format (str): The format of the data source. Defaults to "parquet".
        sequence_column (str): The column name for the peptide sequence in the dataset. Defaults to "sequence_modified".
        label_column (str): The column name for ion mobility in the dataset. Defaults to ["ccs", "ccs_std"].
        val_ratio (float): The ratio of validation data to split from the main dataset. Defaults to 0.2.
        max_seq_len (Union[int, str]): The maximum sequence length allowed in the dataset. Defaults to 30.
        dataset_type (str): The type of dataset to use. Defaults to "tf". Fallback is to TensorFlow dataset tensors.
        batch_size (int): The batch size for the dataset. Defaults to 256.
        model_features (Optional[List[str]]): The features to use in the model. Defaults to ["charge", "mz"].
        dataset_columns_to_keep (Optional[List[str]]): The columns to keep in the dataset. Defaults to None.
        features_to_extract (Optional[List[Union[Callable, str]]]): The features to extract from the dataset. Defaults to None.
        pad (bool): Whether to pad sequences to the maximum length. Defaults to True.
        padding_value (str): The value to use for padding sequences. Defaults to '-'.
        alphabet (Optional[Dict]): The alphabet used for encoding sequences. Defaults to None to trigger learning the alphabet.
        with_termini (bool): Whether to add the N- and C-termini in the sequence column, even if they do not exist. Defaults to True.
        encoding_scheme (Union[str, EncodingScheme]): The encoding scheme to use for sequences. Defaults to EncodingScheme.UNMOD.
        processed (bool): Whether the dataset has been preprocessed. Defaults to False.
        enable_tf_dataset_cache (bool): Flag to indicate whether to enable TensorFlow Dataset caching (call `.cache()` on the generated TF Datasets).
        disable_cache (bool): Whether to disable Hugging Face datasets caching. Default is False.
        num_proc (Optional[int]): Number of processes to use for dataset processing. Use -1 for all available processors, None for single-process mode, or a positive integer. Default is -1.
    """

    def __init__(
        self,
        data_source: Optional[Union[str, List]] = None,
        val_data_source: Optional[Union[str, List]] = None,
        test_data_source: Optional[Union[str, List]] = None,
        data_format: str = "parquet",
        sequence_column: str = "sequence_modified",
        label_column: Union[str, List] = ["ccs", "ccs_std"],
        val_ratio: float = 0.1,
        max_seq_len: Union[int, str] = 50,
        dataset_type: str = "tf",
        batch_size: int = 256,
        model_features: Optional[List[str]] = ["charge", "mz"],
        dataset_columns_to_keep: Optional[List[str]] = None,
        features_to_extract: Optional[List[Union[Callable, str]]] = None,
        pad: bool = True,
        padding_value: str = "-",
        alphabet: Optional[Dict] = None,
        with_termini: bool = True,
        encoding_scheme: Union[str, EncodingScheme] = EncodingScheme.NAIVE_MODS,
        processed: bool = False,
        enable_tf_dataset_cache: bool = False,
        disable_cache: bool = False,
        auto_cleanup_cache: bool = True,
        num_proc: Optional[int] = -1,
        batch_processing_size: int = 1000,
    ):
        kwargs = {k: v for k, v in locals().items() if k not in ["self", "__class__"]}
        super().__init__(DatasetConfig(**kwargs))


def reduced_mobility_to_ccs(
    one_over_k0, mz, charge, mass_gas=28.013, temp=31.85, t_diff=273.15
):
    """
    convert reduced ion mobility (1/k0) to CCS
    :param one_over_k0: reduced ion mobility
    :param charge: charge state of the ion
    :param mz: mass-over-charge of the ion
    :param mass_gas: mass of drift gas
    :param temp: temperature of the drift gas in C°
    :param t_diff: factor to translate from C° to K
    """
    SUMMARY_CONSTANT = 18509.8632163405
    reduced_mass = (mz * charge * mass_gas) / (mz * charge + mass_gas)
    return (SUMMARY_CONSTANT * charge) / (
        np.sqrt(reduced_mass * (temp + t_diff)) * 1 / one_over_k0
    )


def ccs_to_one_over_reduced_mobility(
    ccs, mz, charge, mass_gas=28.013, temp=31.85, t_diff=273.15
):
    """
    convert CCS to 1 over reduced ion mobility (1/k0)
    :param ccs: collision cross-section
    :param charge: charge state of the ion
    :param mz: mass-over-charge of the ion
    :param mass_gas: mass of drift gas (N2)
    :param temp: temperature of the drift gas in C°
    :param t_diff: factor to translate from C° to K
    """
    SUMMARY_CONSTANT = 18509.8632163405
    reduced_mass = (mz * charge * mass_gas) / (mz * charge + mass_gas)
    return ((np.sqrt(reduced_mass * (temp + t_diff))) * ccs) / (
        SUMMARY_CONSTANT * charge
    )


def get_sqrt_weights_and_biases(
    mz: NDArray,
    charge: NDArray,
    ccs: NDArray,
    fit_charge_state_one: bool = True,
    max_charge: int = 4,
) -> Tuple[NDArray, NDArray]:
    """
    Fit a sqrt function to the data and return the weights and biases,
    used to parameterize the init layer for the CCS prediction model.
    Args:
        mz: Array of mass-over-charge values
        charge: Array of charge states
        ccs: Array of collision cross-section values
        fit_charge_state_one: Whether to fit the charge state 1 or not (should be set to false if
        your data does not contain charge state 1)
        max_charge: Maximum charge state to consider

    Returns:
        Tuple of weights and biases the initial projection layer can be parameterized with
    """
    from scipy.optimize import curve_fit

    if fit_charge_state_one:
        slopes, intercepts = [], []
    else:
        slopes, intercepts = [0.0], [0.0]

    c_begin = 1 if fit_charge_state_one else 2

    for c in range(c_begin, max_charge + 1):

        def fit_func(x, a, b):
            return a * np.sqrt(x) + b

        mask = charge == c
        mz_tmp = mz[mask]
        ccs_tmp = ccs[mask]

        popt, _ = curve_fit(fit_func, mz_tmp, ccs_tmp)

        slopes.append(popt[0])
        intercepts.append(popt[1])

    return np.array(slopes, np.float32), np.array(intercepts, np.float32)



import abc
import re
import warnings


class PeptideDatasetBaseProcessor(abc.ABC):
    """
    Base class for peptide dataset processors.

    Parameters
    ----------
    sequence_column_name : str
        Name of the column containing the peptide sequence.
    batched : bool (default=False)
        Whether to process data in batches.
    """

    def __init__(self, sequence_column_name: str = "", batched: bool = False):
        self.sequence_column_name = sequence_column_name
        self.batched = batched

        if self.batched:
            self._process_fn = self.batch_process
        else:
            self._process_fn = self.single_process

    def process(self, input_data, **kwargs):
        return self._process_fn(input_data, **kwargs)

    def __call__(self, input_data, **kwargs):
        return self.process(input_data, **kwargs)

    def __repr__(self):
        members = [
            attr
            for attr in vars(self)
            if not callable(getattr(self, attr)) and not attr.startswith(("_", "__"))
        ]
        values = [self.__getattribute__(m) for m in members]

        repr_str = ", ".join(
            [
                f"{m}='{v}'" if isinstance(v, str) else f"{m}={v}"
                for m, v in zip(members, values)
            ]
        )

        return f"{self.__class__.__name__}({repr_str})"

    @abc.abstractmethod
    def batch_process(self, input_data, **kwargs):
        pass

    @abc.abstractmethod
    def single_process(self, input_data, **kwargs):
        pass


class SequenceParsingProcessor(PeptideDatasetBaseProcessor):
    """
    Processor for parsing peptide sequences in ProForma format.

    Parameters
    ----------
    sequence_column_name : str
        Name of the column containing the peptide sequence.
    batched : bool (default=False)
        Whether to process data in batches.
    with_termini : bool (default=True)
        Whether to add terminal modifications (also in case they do not exists, []- and -[]) to sequence column and overwrite it.

    Attributes
    ----------
    PARSED_COL_NAMES : dict
        Dictionary mapping the parsed sequence parts to new column names.

    Examples
    --------
    >>> processor = SequenceParsingProcessor("sequence")
    >>> data = {"sequence": "[]-IGGPC[UNIMOD:4]AHC[UNIMOD:4]AAWEGVR-[]"}
    >>> processor(data)
    {'sequence': ['[]-', 'I', 'G', 'G', 'P', 'C', 'A', 'H', 'C', 'A', 'A', 'W', 'E', 'G', 'V', 'R', '-[]'],
     '_parsed_sequence': ['I', 'G', 'G', 'P', 'C', 'A', 'H', 'C', 'A', 'A', 'W', 'E', 'G', 'V', 'R'],
     '_n_term_mods': '[]-',
     '_c_term_mods': '-[]'}

    """

    PARSED_COL_NAMES = {
        "seq": "_parsed_sequence",
        "n_term": "_n_term_mods",
        "c_term": "_c_term_mods",
    }

    def __init__(
        self,
        sequence_column_name: str,
        batched: bool = False,
        with_termini: bool = True,
    ):
        super().__init__(sequence_column_name, batched)
        self.with_termini = with_termini

        # decide on sequence update function -> avoid conditional in function/loop
        if self.with_termini:
            self._assign_sequence_column = self.__update_sequence_column_with_termini
        else:
            self._assign_sequence_column = self.__update_sequence_column_without_termini

    def _parse_proforma_sequence(self, sequence_string):
        splitted = sequence_string.split("-")
        seq = ""

        if len(splitted) == 1:
            n_term, seq, c_term = "[]-", splitted[0], "-[]"
        elif len(splitted) == 2:
            if splitted[0].startswith("[UNIMOD:"):
                n_term, seq, c_term = splitted[0] + "-", splitted[1], "-[]"
            else:
                n_term, seq, c_term = "[]-", splitted[0], "-" + splitted[1]
        elif len(splitted) == 3:
            n_term, seq, c_term = splitted
            n_term += "-"
            c_term = "-" + c_term
        else:
            raise ValueError(
                f"Invalid sequence format: {sequence_string}. Expected format: [N-term]-[sequence]-[C-term]."
            )

        seq = re.findall(r"[A-Za-z](?:\[UNIMOD:\d+\])*|[^\[\]]", seq)
        return n_term, seq, c_term

    def __update_sequence_column_with_termini(self, n_terms, seq, c_terms):
        return [n_terms] + seq + [c_terms]

    def __update_sequence_column_without_termini(self, n_terms, seq, c_terms):
        return seq

    def batch_process(self, input_data, **kwargs):
        for new_column in SequenceParsingProcessor.PARSED_COL_NAMES.values():
            input_data[new_column] = []

        for index, sequence in enumerate(input_data[self.sequence_column_name]):
            n_terms, seq, c_terms = self._parse_proforma_sequence(sequence)
            input_data[SequenceParsingProcessor.PARSED_COL_NAMES["seq"]].append(seq)
            input_data[SequenceParsingProcessor.PARSED_COL_NAMES["n_term"]].append(
                n_terms
            )
            input_data[SequenceParsingProcessor.PARSED_COL_NAMES["c_term"]].append(
                c_terms
            )

            # Replace the original sequence with the parsed sequence + terminal mods or parsed sequence only
            input_data[self.sequence_column_name][index] = self._assign_sequence_column(
                n_terms, seq, c_terms
            )

        return input_data

    def single_process(self, input_data, **kwargs):
        n_terms, seq, c_terms = self._parse_proforma_sequence(
            input_data[self.sequence_column_name]
        )
        input_data[SequenceParsingProcessor.PARSED_COL_NAMES["seq"]] = seq
        input_data[SequenceParsingProcessor.PARSED_COL_NAMES["n_term"]] = n_terms
        input_data[SequenceParsingProcessor.PARSED_COL_NAMES["c_term"]] = c_terms

        # Replace the original sequence with the parsed sequence + terminal mods or parsed sequence only
        input_data[self.sequence_column_name] = self._assign_sequence_column(
            n_terms, seq, c_terms
        )

        return input_data


class SequencePaddingProcessor(PeptideDatasetBaseProcessor):
    """
    Processor for padding peptide sequences to a fixed length.

    Parameters
    ----------
    sequence_column_name : str
        Name of the column containing the peptide sequence.
    batched : bool (default=False)
        Whether to process data in batches.
    padding_index : int (default=0)
        Value to use for padding the sequences.
    max_length : int (default=30)
        Maximum length of the sequences.

    Attributes
    ----------
    KEEP_COLUMN_NAME : str
        Name of the column indicating whether to keep the sequence or not (truncated sequences should be dropped from train/val sets).
    """

    KEEP_COLUMN_NAME = "_KEEP"

    def __init__(
        self,
        sequence_column_name: str,
        batched: bool = False,
        padding_index: int = 0,
        max_length: int = 30,
    ):
        super().__init__(sequence_column_name, batched)

        self.padding_index = padding_index
        self.max_length = max_length

    def batch_process(self, input_data, **kwargs):
        input_data[SequencePaddingProcessor.KEEP_COLUMN_NAME] = [True] * len(
            input_data[self.sequence_column_name]
        )

        for index, sequence in enumerate(input_data[self.sequence_column_name]):
            (
                input_data[self.sequence_column_name][index],
                keep_sequence,
            ) = self._pad_sequence(sequence)
            input_data[SequencePaddingProcessor.KEEP_COLUMN_NAME][index] = keep_sequence

        return input_data

    def single_process(self, input_data, **kwargs):
        input_data[self.sequence_column_name], keep_sequence = self._pad_sequence(
            input_data[self.sequence_column_name]
        )

        input_data[SequencePaddingProcessor.KEEP_COLUMN_NAME] = keep_sequence
        return input_data

    def _pad_sequence(self, sequence):
        length = len(sequence)
        if length <= self.max_length:
            return sequence + [self.padding_index] * (self.max_length - length), True
        else:
            return sequence[: self.max_length], False


class SequenceEncodingProcessor(PeptideDatasetBaseProcessor):
    """
    Processor for encoding peptide sequences using an alphabet.

    Parameters
    ----------
    sequence_column_name : str
        Name of the column containing the peptide sequence.
    alphabet : dict
        Dictionary mapping amino acids to integers.
    batched : bool (default=False)
        Whether to process data in batches.
    extend_alphabet : bool (default=False)
        Whether to extend the provided alphabet with new amino acids found in the sequences.
    unknown_token : str (default='X')
        Token to use for unknown amino acids.
    unknown_token_index : int (default=1)
        Index to use for unknown amino acids.
    fallback_unmodified : bool (default=False)
        Whether to fallback to unmodified amino acid encoding for unseen modifications.
    """

    def __init__(
        self,
        sequence_column_name: str,
        alphabet: dict,
        batched: bool = False,
        extend_alphabet: bool = False,
        unknown_token: str = "X",
        unknown_token_index: int = 1,
        fallback_unmodified: bool = False,
    ):
        super().__init__(sequence_column_name, batched)

        self.extend_alphabet = extend_alphabet
        self.unknown_token = unknown_token
        self.unknown_token_index = unknown_token_index

        self.set_alphabet(alphabet)

        # decide on mode, either encode with fixed vocab or learn new tokens
        if self.extend_alphabet:
            self._encode = self._encode_learn_vocab
        else:
            self._encode = self._encode_with_vocab

        self.set_fallback(fallback_unmodified)

    def set_alphabet(self, alphabet):
        # get the passed alphabet from the dataset class, either with the padding only to learn the rest or the full vocabulary
        self.alphabet = alphabet.copy()

        # check if unknown token is already in the alphabet
        unk_token_in_alphabet = self.alphabet.get(self.unknown_token, None)
        if unk_token_in_alphabet:
            warnings.warn(
                f"The unknown token '{self.unknown_token}' is already present in the provided alphabet with index {unk_token_in_alphabet}. If you prefer the default behavior, consider removing it from the alphabet and it will have the default index of {self.unknown_token_index}."
            )
            self.unknown_token_index = unk_token_in_alphabet
        else:
            if self.unknown_token_index in self.alphabet.values():
                new_index = max(self.alphabet.values()) + 1
                warnings.warn(
                    f"The unknown token index '{self.unknown_token_index}' is already used in the provided alphabet. It will be reassigned to index {new_index}."
                )

                self.unknown_token_index = new_index

            warnings.warn(
                f"The unknown token '{self.unknown_token}' is not present in the provided alphabet. It will be added with index {self.unknown_token_index}."
            )
            self.alphabet[self.unknown_token] = self.unknown_token_index

    def set_fallback(self, fallback_unmodified):
        self.fallback_unmodified = fallback_unmodified
        if self.fallback_unmodified:
            self._encode = self._encode_with_vocab_fallback
            if len(self.alphabet) == 0:
                raise ValueError(
                    "Alphabet must be provided if fallback_unmodified is True, to encode unseen modifications with the respective unmodified amino acid token."
                )

    def batch_process(self, input_data, **kwargs):
        return {
            self.sequence_column_name: [
                self._encode(seq) for seq in input_data[self.sequence_column_name]
            ]
        }

    def single_process(self, input_data, **kwargs):
        return {
            self.sequence_column_name: self._encode(
                input_data[self.sequence_column_name]
            )
        }

    def _encode_learn_vocab(self, sequence):
        encoded = []
        for amino_acid in sequence:
            if amino_acid not in self.alphabet:
                self.alphabet[amino_acid] = max(self.alphabet.values()) + 1
            encoded.append(self.alphabet.get(amino_acid))

        return encoded

    def _encode_with_vocab(self, sequence):
        encoded = [
            self.alphabet.get(amino_acid, self.unknown_token_index)
            for amino_acid in sequence
        ]
        return encoded

    def _encode_with_vocab_fallback(self, sequence):
        encoded = []
        for amino_acid in sequence:
            if amino_acid not in self.alphabet:
                if amino_acid.startswith(("[")):
                    amino_acid = "[]-"
                elif amino_acid.startswith(("-[")):
                    amino_acid = "-[]"
                else:
                    amino_acid = amino_acid[0]

            encoded.append(self.alphabet.get(amino_acid, self.unknown_token_index))

        return encoded


class SequencePTMRemovalProcessor(PeptideDatasetBaseProcessor):
    """
    Processor for removing PTMs from peptide sequences.

    Parameters
    ----------
    sequence_column_name : str
        Name of the column containing the peptide sequence.
    batched : bool (default=False)
        Whether to process data in batches.
    """

    def __init__(self, sequence_column_name: str, batched: bool = False):
        super().__init__(sequence_column_name, batched)

    def batch_process(self, input_data, **kwargs):
        return {
            self.sequence_column_name: [
                self._remove_ptms(seq) for seq in input_data[self.sequence_column_name]
            ]
        }

    def single_process(self, input_data, **kwargs):
        return {
            self.sequence_column_name: self._remove_ptms(
                input_data[self.sequence_column_name]
            )
        }

    def _remove_ptms(self, sequence):
        if not isinstance(sequence, list):
            raise ValueError("Sequence must be a list of amino acids")
        n_terms = sequence[0]
        c_terms = sequence[-1]
        aa_sequence = sequence[1:-1]
        ptm_filtered_sequence = re.sub(r"\[UNIMOD:\d+\]", "", "".join(aa_sequence))
        return [n_terms] + list(ptm_filtered_sequence) + [c_terms]


class FunctionProcessor(PeptideDatasetBaseProcessor):
    """
    Processor for applying a function to the input data.

    Parameters
    ----------
    function : callable
        Function to apply to the input data.
    name : str (default="")
        Name of the processor.
    """

    def __init__(self, function, name: str = ""):
        super().__init__(batched=False)
        self.function = function
        self.name = name if name != "" else self.function.__name__

    def batch_process(self, input_data, **kwargs):
        raise NotImplementedError("FunctionProcessor does not support batch processing")

    def single_process(self, input_data, **kwargs):
        return self.function(input_data, **kwargs)



import logging

import torch
from datasets import Dataset

from dlomix.data import FragmentIonIntensityDataset

logger = logging.getLogger(__name__)


def test_dataset_torch(raw_generic_nested_data):
    hfdata = Dataset.from_dict(raw_generic_nested_data)

    intensity_dataset = FragmentIonIntensityDataset(
        data_format="hf",
        data_source=hfdata,
        sequence_column="seq",
        label_column="label",
        model_features=["nested_feature"],
        dataset_type="pt",
        batch_size=2,
        max_seq_len=15,
        with_termini=False,
    )

    logger.info(intensity_dataset)
    assert intensity_dataset.hf_dataset is not None
    assert intensity_dataset._empty_dataset_mode is False

    batch = next(iter(intensity_dataset.tensor_train_data))

    logger.info(batch)

    assert list(batch["nested_feature"].shape) == [2, 1, 2]
    assert list(batch["seq"].shape) == [2, 15]
    assert list(batch["label"].shape) == [
        2,
    ]

    assert batch["seq"].dtype == torch.int64
    assert batch["label"].dtype == torch.float32



__version__ = "0.2.6"
__author__ = "Wilhelm Lab"
__author_email__ = "o.shouman@tum.de"
__license__ = "MIT"
__description__ = "Deep Learning for Proteomics"
__package__ = "DLOmix"
__copyright__ = "2026, Wilhelm Lab, TU Munich, School of Life Sciences"
__github_url__ = "https://github.com/wilhelm-lab/dlomix"



import os
import re

import numpy as np
import pandas as pd
import wandb
import wandb.apis.reports as wr

from ..data.retention_time import RetentionTimeDataset

# ToDo: add R2 plot, TimeDelta plot, residuals


class RetentionTimeReportRunComparisonWandb:
    """Create WandB report for comparing runs.

    Parameters
    ----------
        project (str): Name of the project to be used in wandb.
        title (str): Title of the report in wandb.
        description (str): Description of the report in wandb.
        dataset (RetentionTimeDataset, optional): The retention time dataset if logging the data is desired. Defaults to None, no logging of input data.
    """

    METRICS_TO_EXCLUDE = [
        "epoch/learning_rate",
        "epoch/epoch",
        "batch/learning_rate",
        "batch/batch_step",
    ]

    # Wilhelmlab WandB account that has all VEGA presets required for the reports
    VEGA_LITE_PRESETS_ID = "prosit-compms"

    def __init__(
        self,
        project: str,
        title: str,
        description: str,
        dataset: RetentionTimeDataset = None,
    ):
        self.project = project
        self.title = title
        self.description = description
        self.dataset = dataset
        self.entity = wandb.apis.PublicApi().default_entity
        self.wandb_api = wandb.Api()
        self.table_key_len = ""
        self.table_key_rt = ""
        self.model_info = []

    def create_report(
        self,
        add_config_section=True,
        add_data_section=True,
        add_train_section=True,
        add_val_section=True,
        add_train_val_section=True,
        add_model_section=True,
    ):
        """Create a report in wandb.

        Args:
            add_config_section (bool, optional): Add a section for config parameters and the run to the report. Defaults to True.
            add_data_section (bool, optional): Add a section for input data to the report. Defaults to True.
            add_train_section (bool, optional): Add a section for training metrics to the report. Defaults to True.
            add_val_section (bool, optional): Add a section for validation metrics to the report. Defaults to True.
            add_train_val_section (bool, optional): Add a section for train-val metrics to the report. Defaults to True.
            add_model_section (bool, optional): Add a section for model summary and number of parameters to the report. Defaults to True.
        """
        report = wr.Report(
            project=self.project, title=self.title, description=self.description
        )

        report.blocks = [wr.TableOfContents()]

        if add_model_section:
            report.blocks += self._build_model_section()
        if add_config_section:
            report.blocks += self._build_config_section()
        if add_data_section and self.dataset is not None:
            report.blocks += self._build_data_section()
        if add_train_section:
            report.blocks += self._build_train_section()
        if add_val_section:
            report.blocks += self._build_val_section()
        if add_train_val_section:
            report.blocks += self._build_train_val_section()

        report.save()

    # get metrics of last run in project or from specified run_id
    def _get_metrics(self, run_id=None):
        if run_id:
            # run is specified by <entity>/<project>/<run_id>
            run = self.wandb_api.run(path=f"{self.entity}/{self.project}/{run_id}")
            metrics_dataframe = run.history()
            return metrics_dataframe
        else:
            # get metrics of latest run
            runs = self.wandb_api.runs(path=f"{self.entity}/{self.project}")
            run = runs[0]
            metrics_dataframe = run.history()
            return metrics_dataframe

    # get metric names split into train/val, train is further split into batch/epoch
    def _get_metrics_names(self):
        metrics = self._get_metrics()
        # filter strings from list that are not starting with "_" and do not contain "val"
        pre_filter = [string for string in metrics if not string.startswith("_")]
        batch_train_metrics_names = [
            string
            for string in pre_filter
            if ("val" not in string.lower())
            & ("epoch" not in string.lower())
            & ("table" not in string.lower())
        ]
        epoch_train_metrics_names = [
            string
            for string in pre_filter
            if ("val" not in string.lower())
            & ("batch" not in string.lower())
            & ("table" not in string.lower())
        ]
        # filter strings from list that contain "val"
        epoch_val_metrics_names = list(filter(lambda x: "val" in x.lower(), metrics))

        # filter strings from train metrics that are 'epoch/learning_rate' and 'epoch/epoch'
        strings_to_filter = RetentionTimeReportRunComparisonWandb.METRICS_TO_EXCLUDE
        batch_train_metrics_names = [
            string
            for string in batch_train_metrics_names
            if string not in strings_to_filter
        ]
        epoch_train_metrics_names = [
            string
            for string in epoch_train_metrics_names
            if string not in strings_to_filter
        ]
        batch_train_metrics_names.sort()
        epoch_train_metrics_names.sort()
        return (
            batch_train_metrics_names,
            epoch_train_metrics_names,
            epoch_val_metrics_names,
        )

    def get_train_val_metrics_names(self):
        (
            _,
            epoch_train_metrics_names,
            epoch_val_metrics_names,
        ) = self._get_metrics_names()
        epoch_train_metrics_names.sort()
        epoch_val_metrics_names.sort()
        return list(zip(epoch_train_metrics_names, epoch_val_metrics_names))

    def _build_config_section(self):
        config_block = [
            wr.H1(text="Config"),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=[wr.RunComparer(layout={"w": 24})],
            ),
            wr.HorizontalRule(),
        ]
        return config_block

    def _build_data_section(self):
        data_block = [
            wr.H1(text="Data"),
            wr.P(
                "The following section is showing a simple explorative data analysis of the used dataset. The first histogram shows the distribution of peptide lengths in the data set, while the second histogram shows the distribution of indexed retention times."
            ),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=[
                    wr.CustomChart(
                        query={"summaryTable": {"tableKey": self.table_key_len}},
                        chart_name=f"{RetentionTimeReportRunComparisonWandb.VEGA_LITE_PRESETS_ID}/histogram_peptide_length",
                        chart_fields={"value": self.dataset.sequence_col},
                    ),
                    wr.CustomChart(
                        query={"summaryTable": {"tableKey": self.table_key_rt}},
                        chart_name=f"{RetentionTimeReportRunComparisonWandb.VEGA_LITE_PRESETS_ID}/histogram_irt",
                        chart_fields={"value": self.dataset.target_col},
                    ),
                ],
            ),
            wr.HorizontalRule(),
        ]
        return data_block

    def _build_train_section(self):
        (
            batch_train_metrics_names,
            epoch_train_metrics_names,
            _,
        ) = self._get_metrics_names()
        panel_list_batch = []
        panel_list_epoch = []
        for name in batch_train_metrics_names:
            panel_list_batch.append(wr.LinePlot(x="Step", y=[name]))
        for name in epoch_train_metrics_names:
            panel_list_epoch.append(wr.LinePlot(x="Step", y=[name]))
        train_block = [
            wr.H1(text="Training metrics"),
            wr.P(
                "The following section shows the different metrics that were used to track the training. All used metrics are added by default. The first subsection shows the metrics per epoch, whereas the second subsection show the metrics per batch."
            ),
            wr.H2(text="per batch"),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=panel_list_batch,
            ),
            wr.H2(text="per epoch"),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=panel_list_epoch,
            ),
            wr.HorizontalRule(),
        ]
        return train_block

    def _build_val_section(self):
        _, _, epoch_val_metrics_names = self._get_metrics_names()
        panel_list_epoch = []
        for name in epoch_val_metrics_names:
            panel_list_epoch.append(wr.LinePlot(x="Step", y=[name]))
        val_block = [
            wr.H1(text="Validation metrics"),
            wr.P(
                "The following section shows the different metrics that were used to track the validation. All used metrics are added by default. The metrics are shown per epoch."
            ),
            wr.H2(text="per epoch"),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=panel_list_epoch,
            ),
            wr.HorizontalRule(),
        ]
        return val_block

    def _build_model_section(self):
        model_block = [
            wr.H1(text="Model information"),
            wr.P(
                "The following section shows information about the model. The table below contains information about the models' layers."
            ),
            wr.UnorderedList(items=self.model_info),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=[wr.WeavePanelSummaryTable("layer_table")],
            ),
            wr.HorizontalRule(),
        ]
        return model_block

    def _build_train_val_section(self):
        train_val_metrics_names = self.get_train_val_metrics_names()
        panel_list_epoch = []
        for name in train_val_metrics_names:
            panel_list_epoch.append(wr.LinePlot(x="Step", y=list(name)))
        train_val_block = [
            wr.H1(text="Train - Validation metrics"),
            wr.P(
                "The following section shows the different metrics for both training and validation in comaprison. All used metrics are added by default. The metrics are shown per epoch."
            ),
            wr.H2(text="per epoch"),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=panel_list_epoch,
            ),
            wr.HorizontalRule(),
        ]
        return train_val_block

    def log_sequence_length_table(
        self, data: pd.DataFrame, seq_col: str = "modified_sequence"
    ):
        """Log sequence length table to wandb

        Args:
            data (pd.DataFrame): input data
            seq_col (str, optional): Name of the column containing the sequences in the data frame. Defaults to "modified_sequence".

        Returns:
            str: Name of the histogram created by wandb after logging the data.
        """
        name_hist = "counts_hist"
        counts = self.count_seq_length(data, seq_col)
        # convert to df for easier handling
        counts_df = counts.to_frame()
        table = wandb.Table(dataframe=counts_df)
        # log to wandb
        hist = wandb.plot_table(
            vega_spec_name=f"{RetentionTimeReportRunComparisonWandb.VEGA_LITE_PRESETS_ID}/histogram_peptide_length",
            data_table=table,
            fields={"value": seq_col},
        )
        wandb.log({name_hist: hist})
        name_hist_table = name_hist + "_table"
        return name_hist_table

    # function to count sequence length
    def count_seq_length(self, data: pd.DataFrame, seq_col: str):
        pattern = re.compile(r"\[UNIMOD:.*\]", re.IGNORECASE)
        data[seq_col].replace(pattern, "", inplace=True)
        return data[seq_col].str.len()

    # function to log retention time table to wandb
    def log_rt_table(self, data: pd.DataFrame, rt_col: str = "indexed_retention_time"):
        name_hist = "rt_hist"
        rt = data.loc[:, rt_col]
        # convert to df for easier handling
        rt_df = rt.to_frame()
        table = wandb.Table(dataframe=rt_df)
        # log to wandb
        hist = wandb.plot_table(
            vega_spec_name=f"{RetentionTimeReportRunComparisonWandb.VEGA_LITE_PRESETS_ID}/histogram_irt",
            data_table=table,
            fields={"value": rt_col},
        )
        wandb.log({name_hist: hist})
        name_hist_table = name_hist + "_table"
        return name_hist_table

    def log_data(self):
        # check if datasource is a string
        if isinstance(self.dataset.data_source, str):
            # read corresponding file
            file_extension = os.path.splitext(self.dataset.data_source)[-1].lower()
            data = pd.DataFrame()

            if file_extension == ".csv":
                data = pd.read_csv(self.dataset.data_source)
            if file_extension == ".json":
                data = pd.read_json(self.dataset.data_source)
            if file_extension == ".parquet":
                data = pd.read_parquet(self.dataset.data_source, engine="fastparquet")
            self.table_key_len = self.log_sequence_length_table(
                data, self.dataset.sequence_col
            )
            self.table_key_rt = self.log_rt_table(data, self.dataset.target_col)

        # check if datasource is a tuple of two ndarrays or two lists
        if (
            isinstance(self.dataset.data_source, tuple)
            and all(
                isinstance(item, (np.ndarray, list))
                for item in self.dataset.data_source
            )
            and len(self.dataset.data_source) == 2
        ):
            data = pd.DataFrame(
                {
                    self.dataset.sequence_col: self.dataset.data_source[0],
                    self.dataset.target_col: self.dataset.data_source[1],
                }
            )
            self.table_key_len = self.log_sequence_length_table(
                data, self.dataset.sequence_col
            )
            self.table_key_rt = self.log_rt_table(data, self.dataset.target_col)

        # check if datasource is a single ndarray or list
        # does not work? maybe error in RetentionTimeDataset
        if isinstance(self.dataset.data_source, (np.ndarray, list)):
            data = pd.DataFrame({self.dataset.sequence_col: self.dataset.data_source})
            self.table_key_len = self.log_sequence_length_table(
                data, self.dataset.sequence_col
            )

    def log_model_data(self, model):
        import io

        model_summary_buffer = io.StringIO()
        model.summary(print_fn=lambda x: model_summary_buffer.write(x + "<br>"))
        model_summary_lines = model_summary_buffer.getvalue().split("<br>")

        lines = [line.rstrip() for line in model_summary_lines]

        # remove formatting lines
        strings_to_remove = ["____", "===="]
        cleaned_list = [
            item
            for item in lines
            if not any(string in item for string in strings_to_remove)
        ]

        # split into words by splitting if there are more than two whitespaces
        words = []
        for line in cleaned_list:
            words.append(re.split(r"\s{2,}", line))

        # remove lines that contain less than 3 characters
        filtered_list_of_lists = [
            sublist for sublist in words if all(len(item) > 3 for item in sublist)
        ]

        # extract layer info and model info
        layer_info = [sublist for sublist in filtered_list_of_lists if len(sublist) > 2]
        model_info = [sublist for sublist in filtered_list_of_lists if len(sublist) < 2]

        # flatten model_info and filter entries with length smaller than 5
        model_info_flat = [item for sublist in model_info for item in sublist]
        model_info_flat_filtered = [item for item in model_info_flat if len(item) >= 5]

        # create layer_info_df
        column_names = layer_info[0]
        layer_info_df = pd.DataFrame(layer_info[1:], columns=column_names)

        # log layer_table to wandb
        layer_table = wandb.Table(dataframe=layer_info_df)
        wandb.log({"layer_table": layer_table})

        # attach model_info to object
        self.model_info = model_info_flat_filtered



from os.path import join
from warnings import warn

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.ticker import LogLocator

from ..reports.Report import PDFFile, Report


class RetentionTimeReport(Report):
    """Report generation for Retention Time Prediction tasks."""

    TARGETS_LABEL = "iRT (measured)"
    PREDICTIONS_LABEL = "iRT (predicted)"

    def __init__(self, output_path, history, figures_ext="png"):
        super(RetentionTimeReport, self).__init__(output_path, history, figures_ext)

        warn(
            f"{self.__class__.__name__} This class is deprecated and will not further developed. Use RetentionTimeReportWandb instead for creating a report with the Weights & Biases Report API.",
            DeprecationWarning,
            stacklevel=2,
        )

        self.pdf_file = PDFFile("DLOmix - Retention Time Report")

    def generate_report(self, targets, predictions, **kwargs):
        self._init_report_resources()

        r2 = self.calculate_r2(targets, predictions)
        self.plot_all_metrics()
        self.plot_residuals(targets, predictions)
        self.plot_density(targets, predictions)

        self._compile_report_resources_add_pdf_pages()

        self.pdf_file.output(join(self._output_path, "iRT_Report.pdf"), "F")

    def calculate_r2(self, targets, predictions):
        """Calculate R-squared using sklearn given true targets and predictions

        Arguments
        ---------
            targets: Array with target values
            predictions: Array with prediction values

        Returns:
            r_squared (float): float value of R squared
        """
        from sklearn.metrics import r2_score

        r2 = r2_score(np.ravel(targets), np.ravel(predictions))

        self._add_report_resource(
            "r2",
            "R-Squared",
            f"The R-squared value for the predictions is {round(r2, 4)}",
            r2,
        )

        return r2

    def plot_residuals(self, targets, predictions, xrange=(0, 0)):
        """Plot histogram of residuals

        Argsuments
        ----------
            targets: Array with target values
            predictions: Array with prediction values
            xrange (tuple, optional): X-axis range for plotting the histogram. Defaults to (-10, 10).
        """
        error = np.ravel(targets) - np.ravel(predictions)

        x_min, x_max = xrange
        if xrange == (0, 0):
            mean, std_dev = np.mean(error), np.std(error)
            x_min, x_max = mean - (3 * std_dev), mean + (3 * std_dev)

        bins = np.linspace(x_min, x_max, 200)

        plt.hist(error, bins, alpha=0.5, color="orange")
        plt.title("Historgram of Residuals")
        plt.xlabel("Residual value")
        plt.ylabel("Count")
        save_path = join(self._output_path, "histogram_residuals" + self._figures_ext)
        plt.savefig(save_path)
        plt.show()
        plt.close()

        self._add_report_resource(
            "residuals_plot",
            "Error Residuals",
            "The following plot shows a historgram of residuals for the test data.",
            save_path,
        )

    def plot_density(
        self,
        targets,
        predictions,
        irt_delta95=5,
        palette="Reds_r",
        delta95_line_color="#36479E",
        nbins=1000,
    ):
        """Create density plot

        Arguments
        ---------
            targets:  Array with target values
            predictions:  Array with prediction values
            irt_delta95 (int, optional): iRT Value of the delta 95% . Defaults to 5.
            palette (str, optional): Color palette from matplotlib. Defaults to 'Reds_r'.
            delta95_line_color (str, optional): Color for the delta 95% line. Defaults to '#36479E'.
            nbins (int, optional): Number of bins to use for creating the 2D histogram. Defaults to 1000.
        """

        H, xedges, yedges = np.histogram2d(targets, predictions, bins=nbins)

        x_min = np.min(targets)
        x_max = np.max(targets)

        # H needs to be rotated and flipped
        H = np.rot90(H)
        H = np.flipud(H)

        # Mask zeros
        Hmasked = np.ma.masked_where(H == 0, H)  # Mask pixels with a value of zero

        # Plot 2D histogram using pcolor
        cm = plt.cm.get_cmap(palette)
        plt.pcolormesh(
            xedges, yedges, Hmasked, cmap=cm, norm=LogNorm(vmin=1e0, vmax=1e2)
        )

        plt.xlabel(RetentionTimeReport.TARGETS_LABEL, fontsize=18)
        plt.ylabel(RetentionTimeReport.PREDICTIONS_LABEL, fontsize=18)

        cbar = plt.colorbar(ticks=LogLocator(subs=range(5)))
        cbar.ax.set_ylabel("Counts", fontsize=14)

        plt.plot([x_min, x_max], [x_min, x_max], c="black")
        plt.plot(
            [x_min, x_max],
            [x_min - irt_delta95, x_max - irt_delta95],
            color=delta95_line_color,
        )
        plt.plot(
            [x_min, x_max],
            [x_min + irt_delta95, x_max + irt_delta95],
            color=delta95_line_color,
        )

        font_size = 14  # Adjust as appropriate.
        cbar.ax.tick_params(labelsize=font_size)
        cbar.ax.minorticks_on()
        save_path = join(self._output_path, "density_plot" + self._figures_ext)
        plt.savefig(save_path)
        plt.show()
        plt.close()

        self._add_report_resource(
            "density_plot",
            "Density Plot",
            "The following figure shows the density plot with the delta-95 highlighted for the test data.",
            save_path,
        )



import functools

import numpy as np
import tensorflow as tf

from ..losses import masked_spectral_distance


def reshape_dims(array):
    n, dims = array.shape
    assert dims == 174
    nlosses = 1
    return array.reshape([array.shape[0], 30 - 1, 2, nlosses, 3])


def reshape_flat(array):
    s = array.shape
    flat_dim = [s[0], functools.reduce(lambda x, y: x * y, s[1:], 1)]
    return array.reshape(flat_dim)


def normalize_base_peak(array):
    # flat
    maxima = array.max(axis=1)
    array = array / maxima[:, np.newaxis]
    return array


def mask_outofrange(array, lengths, mask=-1.0):
    # dim
    for i in range(array.shape[0]):
        array[i, lengths[i] - 1 :, :, :, :] = mask
    return array


def mask_outofcharge(array, charges, mask=-1.0):
    # dim
    for i in range(array.shape[0]):
        if charges[i] < 3:
            array[i, :, :, :, charges[i] :] = mask
    return array


def calculate_spectral_angle(true, pred, batch_size=600):
    """

    Calculates masked spectral distance normalized to [0, 1] range.
    Values of -1 in true are treated as masked (missing peaks).

    Args:
        true: Ground truth array of shape (n, num_values) or array of arrays
        pred: Predicted array of shape (n, num_values) or array of arrays
        batch_size: Kept for API compatibility with old tensorflow 1 version (not used)

    Returns:
        Array of shape (n,) with spectral angles in range [0, 1]
    """
    # Handle pandas Series with array values (common case)
    if hasattr(true, "values"):
        true = true.values
    if hasattr(pred, "values"):
        pred = pred.values

    # Convert array of arrays to 2D array if needed
    if true.dtype == object:
        true = np.stack(true)
    if pred.dtype == object:
        pred = np.stack(pred)

    # Ensure proper float dtype
    true = np.asarray(true, dtype=np.float64)
    pred = np.asarray(pred, dtype=np.float64)

    n = true.shape[0]
    epsilon = 1e-7  # Matches K.epsilon() default

    # Masking: multiply by (true + 1) to mask out -1 values
    pred_masked = ((true + 1) * pred) / (true + 1 + epsilon)
    true_masked = ((true + 1) * true) / (true + 1 + epsilon)

    # L2 normalization (note: original has swapped variables, reproducing exactly)
    # Original: pred_norm = K.l2_normalize(true_masked, axis=-1)
    #           true_norm = K.l2_normalize(pred_masked, axis=-1)
    pred_norm = true_masked / (
        np.linalg.norm(true_masked, axis=-1, keepdims=True) + epsilon
    )
    true_norm = pred_masked / (
        np.linalg.norm(pred_masked, axis=-1, keepdims=True) + epsilon
    )

    # Spectral angle calculation
    product = np.sum(pred_norm * true_norm, axis=1)
    product = np.clip(product, -1.0, 1.0)  # Numerical stability
    arccos = np.arccos(product)

    # Normalize to [0, 1] range
    spectral_distance = 2 * arccos / np.pi

    # Original code does: sa = 1 - s.run(sa_graph)
    # So we invert it here to match exactly
    sa = 1 - spectral_distance

    # Handle NaN values
    sa = np.nan_to_num(sa)

    return sa


def get_spectral_angle(true, pred, batch_size=600):
    """Legacy sepctral angle calculation using TensorFlow 1.x session"""
    n = true.shape[0]
    sa = np.zeros([n])

    def iterate():
        if n > batch_size:
            for i in range(n // batch_size):
                true_sample = true[i * batch_size : (i + 1) * batch_size]
                pred_sample = pred[i * batch_size : (i + 1) * batch_size]
                yield i, true_sample, pred_sample
            i = n // batch_size
            yield i, true[(i) * batch_size :], pred[(i) * batch_size :]
        else:
            yield 0, true, pred

    for i, t_b, p_b in iterate():
        tf.compat.v1.reset_default_graph()
        with tf.compat.v1.Session() as s:
            sa_graph = masked_spectral_distance(t_b, p_b)
            sa_b = 1 - s.run(sa_graph)
            sa[i * batch_size : i * batch_size + sa_b.shape[0]] = sa_b
    sa = np.nan_to_num(sa)
    return sa


def normalize_intensity_predictions(
    data,
    sequence_column_name="sequences",
    labels_column_name="intensities_raw",
    predictions_column_name="intensities_pred",
    precursor_charge_column_name="precursor_charge_onehot",
    batch_size=600,
    compute_spectral_angle=True,
    use_legacy_tf_sa_fn=False,
):
    assert (
        sequence_column_name in data
    ), "Key sequences is missing in the data provided for post-processing"
    assert (
        predictions_column_name in data
    ), "Key intensities_pred is missing in the data provided for post-processing"
    assert (
        precursor_charge_column_name in data
    ), "Key precursor_charge_onehot is missing in the data provided for post-processing"

    sequence_lengths = data[sequence_column_name].apply(lambda x: len(x))
    intensities = np.stack(data[predictions_column_name].to_numpy()).astype(np.float32)
    precursor_charge_onehot = np.stack(data[precursor_charge_column_name].to_numpy())
    charges = list(precursor_charge_onehot.argmax(axis=1) + 1)

    intensities[intensities < 0] = 0
    intensities = reshape_dims(intensities)
    intensities = mask_outofrange(intensities, sequence_lengths)
    intensities = mask_outofcharge(intensities, charges)
    intensities = reshape_flat(intensities)
    m_idx = intensities == -1
    intensities = normalize_base_peak(intensities)
    intensities[m_idx] = -1
    data[predictions_column_name] = intensities.tolist()

    if labels_column_name in data and compute_spectral_angle:
        if use_legacy_tf_sa_fn:
            data["spectral_angle"] = get_spectral_angle(
                np.stack(data[labels_column_name].to_numpy()).astype(np.float32),
                intensities,
                batch_size=batch_size,
            )
        else:
            data["spectral_angle"] = calculate_spectral_angle(
                np.stack(data[labels_column_name].to_numpy()).astype(np.float32),
                intensities,
                batch_size=batch_size,
            )
    return data



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


def get_metadata():
    metadata = {}
    with open("src/dlomix/_metadata.py") as f:
        exec(f.read(), metadata)
    return metadata


# Load metadata
META_DATA = get_metadata()

tensorflow_extra_install = [
    "tensorflow>=2.13,<2.16",  # 2.16 introduces breaking changes and has Keras 3 as default
]

pytorch_extra_install = [
    "torch",
    "torchvision",
]

setuptools.setup(
    name=META_DATA["__package__"].lower(),
    version=META_DATA["__version__"],
    author=META_DATA["__author__"],
    author_email=META_DATA["__author_email__"],
    description=META_DATA["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=META_DATA["__github_url__"],
    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    package_data={"": ["data/processing/pickled_feature_dicts/*"]},
    python_requires=">=3.10",
    install_requires=[
        "datasets>=4.0.0",
        "huggingface_hub",
        "fpdf",
        "pandas",
        "numpy",
        "matplotlib",
        "scikit-learn",
        "pyarrow",
        "seaborn",
    ],
    extras_require={
        "dev": [
            "pytest >= 7.0.0",
            "pytest-cov",
            "black",
            "twine",
            "setuptools",
            "wheel",
            "pylint",
            *tensorflow_extra_install,
            *pytorch_extra_install,
        ],
        "wandb": [
            "wandb>=0.20.0",
        ],
        "tensorflow": tensorflow_extra_install,
        "tf": tensorflow_extra_install,
        "torch": pytorch_extra_install,
        "pytorch": pytorch_extra_install,
        "lightning": [
            "lightning",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research",
    ],
)



import numpy as np
import torch
import torch.nn as nn


class SquareRootProjectionLayer(nn.Module):
    def __init__(self, weights, bias, trainable=True):
        """
        Initialize the SquareRootProjectionLayer.
        Args:
            weights: array of weights
            bias: array of biases
            trainable: whether the weights and biases are optimized during training
        """
        super(SquareRootProjectionLayer, self).__init__()
        self.trainable = trainable
        weights = torch.tensor(weights, dtype=torch.float32)
        bias = torch.tensor(bias, dtype=torch.float32)
        if self.trainable:
            self.slopes = nn.Parameter(weights)
            self.intercepts = nn.Parameter(bias)
        else:
            self.register_buffer("weights", weights)
            self.register_buffer("bias", bias)

    def forward(self, mz, charge):
        sqrt_mz = torch.sqrt(mz)
        sqrt_mz = sqrt_mz.expand(-1, self.slopes.shape[0])
        projection = (self.slopes * sqrt_mz) + self.intercepts
        projection = projection * charge
        result = torch.sum(projection, dim=-1, keepdim=True)
        return result


class Ionmob(nn.Module):
    def __init__(
        self,
        num_tokens,
        initial_weights: np.ndarray = np.array([12.3177, 15.0300, 17.1686, 21.1792]),
        initial_bias: np.ndarray = np.array([-81.5547, 1.8667, 99.4165, 180.1543]),
        max_charge: int = 4,
        max_peptide_length: int = 50,
        emb_dim: int = 64,
        gru_1: int = 64,
        gru_2: int = 32,
        rdo: float = 0.0,
        do: float = 0.2,
    ):
        """
        Initialize the Ionmob model for CCS mean and std prediction.
        Args:
            initial_weights: initial fit weights for the square root projection layer
            initial_bias: initial fit bias(es) for the square root projection layer
            num_tokens: size of the token vocabulary
            max_peptide_length: maximum peptide length (number of amino acids WITH modifications)
            emb_dim: embedding dimension size
            gru_1: size of the first GRU layer
            gru_2: size of the second GRU layer
            rdo: recurrent dropout rate
            do: dropout rate
        """
        super(Ionmob, self).__init__()
        self.max_charge = max_charge
        self.max_peptide_length = max_peptide_length
        self.initial = SquareRootProjectionLayer(
            initial_weights, initial_bias, trainable=True
        )
        self.emb = nn.Embedding(num_tokens, emb_dim)
        self.gru1 = nn.GRU(
            emb_dim, gru_1, batch_first=True, bidirectional=True, dropout=rdo
        )
        self.gru2 = nn.GRU(
            gru_1 * 2, gru_2, batch_first=True, bidirectional=True, dropout=rdo
        )
        self.dropout = nn.Dropout(do)

        # The dense layer input size is the size of the
        # second GRU layer * 2 (bidirectional) + max_charge
        dense1_input_size = gru_2 * 2 + max_charge
        self.dense_ccs_1 = nn.Linear(dense1_input_size, 128)
        self.dense_ccs_2 = nn.Linear(128, 64)

        self.dense_ccs_std_1 = nn.Linear(dense1_input_size, 128)
        self.dense_ccs_std_2 = nn.Linear(128, 64)

        self.out_ccs = nn.Linear(64, 1)
        self.out_ccs_std = nn.Linear(64, 1)

        self.relu = nn.ReLU()

    def forward(self, seq, mz, charge):
        """
        Forward pass of the Ionmob model.
        Args:
            seq: Tokenized peptide sequence
            mz: Mass-over-charge (monoisotopic)
            charge: Charge state of the ion

        Returns:
            total_output: initial sqrt prediction + deep learning prediction
            ccs_output: deep learning output for CCS (difference from initial prediction)
            ccs_std_output: CCS std prediction
        """

        # Ensure seq is of type torch.long
        seq = seq.long()
        batch_size = mz.size(0)
        x_emb = self.emb(seq)

        # one-hot encode charge
        charge = torch.nn.functional.one_hot(
            charge - 1, num_classes=self.max_charge
        ).float()

        # check if mz is (batch_, 1) and not (batch_,), otherwise expand
        if mz.dim() == 1:
            mz = mz.unsqueeze(1)

        x_gru1, _ = self.gru1(x_emb)
        x_gru2, h_n = self.gru2(x_gru1)

        # Get the last hidden state from both directions
        x_recurrent = h_n.permute(1, 0, 2).reshape(batch_size, -1)

        # Concatenate charge and recurrent features
        concat = torch.cat([charge, x_recurrent], dim=-1)

        # Dense layers with activation and dropout
        cc1 = self.dropout(self.relu(self.dense_ccs_1(concat)))
        cc2 = self.relu(self.dense_ccs_2(cc1))

        ccs_std1 = self.dropout(self.relu(self.dense_ccs_std_1(concat)))
        ccs_std2 = self.relu(self.dense_ccs_std_2(ccs_std1))

        # Outputs
        initial_output = self.initial(mz, charge)
        ccs_output = self.out_ccs(cc2)

        ccs_std_output = self.out_ccs_std(ccs_std2)
        total_output = initial_output + ccs_output

        return total_output, ccs_output, ccs_std_output



# most of this can go away, as it is not used anymore

from enum import Enum


class EncodingScheme(str, Enum):
    """
    Enum for encoding schemes.
    """

    UNMOD = "unmod"
    NAIVE_MODS = "naive-mods"

    @classmethod
    def _missing_(cls, value):
        raise ValueError(
            f"Provided {cls.__name__} is not defined. Valid values are: {', '.join([e.value for e in cls])}"
        )


def get_num_processors():
    """
    Get the number of processors available on the system.

    Returns
    -------
    int
        Number of processors available.
    """

    import multiprocessing

    return multiprocessing.cpu_count()


def validate_num_proc_value(num_proc):
    """
    Validate the user-provided num_proc value.

    Allowed values:
    - None: force single-process execution
    - -1: use all available processors
    - positive integer: explicit number of processors
    """

    if num_proc is None:
        return

    if not isinstance(num_proc, int):
        raise ValueError(
            "The num_proc parameter should be None, -1, or a positive integer."
        )

    if num_proc != -1 and num_proc <= 0:
        raise ValueError(
            "Invalid num_proc value. Use None for single-process mode, -1 for all available processors, or a positive integer."
        )


def resolve_num_proc(num_proc, n_processors):
    """
    Resolve num_proc to the effective value used by HF Dataset map/filter calls.

    Returns
    -------
    tuple
        (resolved_num_proc, was_capped)
    """

    validate_num_proc_value(num_proc)

    if num_proc is None:
        return None, False

    if num_proc == -1:
        return n_processors, False

    if num_proc > n_processors:
        return n_processors, True

    return num_proc, False



import tensorflow as tf

from ..constants import CLASSES_LABELS, padding_char


@tf.keras.utils.register_keras_serializable(package="dlomix")
class DetectabilityModel(tf.keras.Model):
    def __init__(
        self,
        num_units,
        num_classes=len(CLASSES_LABELS),
        name="autoencoder",
        padding_char=padding_char,
        **kwargs
    ):
        super(DetectabilityModel, self).__init__(name=name, **kwargs)

        self.num_units = num_units
        self.num_classes = num_classes
        self.padding_char = padding_char
        self.alphabet_size = len(padding_char)
        self.one_hot_encoder = tf.keras.layers.Lambda(
            lambda x: tf.one_hot(tf.cast(x, "int32"), depth=self.alphabet_size)
        )
        self.encoder = Encoder(self.num_units)
        self.decoder = Decoder(self.num_units, self.num_classes)

    def call(self, inputs):
        onehot_inputs = self.one_hot_encoder(inputs)
        enc_outputs, enc_state_f, enc_state_b = self.encoder(onehot_inputs)

        dec_outputs = tf.concat([enc_state_f, enc_state_b], axis=-1)

        decoder_inputs = {
            "decoder_outputs": dec_outputs,
            "state_f": enc_state_f,
            "state_b": enc_state_b,
            "encoder_outputs": enc_outputs,
        }

        decoder_output = self.decoder(decoder_inputs)

        return decoder_output

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_units": self.num_units,
                "num_classes": self.num_classes,
                "padding_char": self.padding_char,
            }
        )
        return config


@tf.keras.utils.register_keras_serializable(package="dlomix")
class Encoder(tf.keras.layers.Layer):
    def __init__(self, units, name="encoder", **kwargs):
        super(Encoder, self).__init__(name=name, **kwargs)

        self.units = units

        self.mask_enco = tf.keras.layers.Masking(mask_value=padding_char)

        self.encoder_gru = tf.keras.layers.GRU(
            self.units,
            return_sequences=True,
            return_state=True,
            recurrent_initializer="glorot_uniform",
        )

        self.encoder_bi = tf.keras.layers.Bidirectional(self.encoder_gru)

    def call(self, inputs):
        mask_ = self.mask_enco.compute_mask(inputs)

        mask_bi = self.encoder_bi.compute_mask(inputs, mask_)

        encoder_outputs, encoder_state_f, encoder_state_b = self.encoder_bi(
            inputs, initial_state=None, mask=mask_bi
        )

        return encoder_outputs, encoder_state_f, encoder_state_b

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "units": self.units,
            }
        )
        return config


@tf.keras.utils.register_keras_serializable(package="dlomix")
class BahdanauAttention(tf.keras.layers.Layer):
    def __init__(self, units, name="attention_layer", **kwargs):
        super(BahdanauAttention, self).__init__(name=name, **kwargs)
        self.units = units
        self.W1 = tf.keras.layers.Dense(units)
        self.W2 = tf.keras.layers.Dense(units)
        self.V = tf.keras.layers.Dense(1)

    def call(self, inputs):
        query = inputs["query"]
        values = inputs["values"]

        query_with_time_axis = tf.expand_dims(query, axis=1)

        query_values = tf.nn.tanh(self.W1(query_with_time_axis) + self.W2(values))
        scores = self.V(query_values)

        attention_weights = tf.nn.softmax(scores, axis=1)

        context_vector = attention_weights * values

        context_vector = tf.reduce_sum(context_vector, axis=1)

        return context_vector

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "units": self.units,
            }
        )
        return config


@tf.keras.utils.register_keras_serializable(package="dlomix")
class Decoder(tf.keras.layers.Layer):
    def __init__(self, units, num_classes, name="decoder", **kwargs):
        super(Decoder, self).__init__(name=name, **kwargs)
        self.units = units
        self.num_classes = num_classes

        self.decoder_gru = tf.keras.layers.GRU(
            self.units, return_state=True, recurrent_initializer="glorot_uniform"
        )

        self.attention = BahdanauAttention(self.units)

        self.decoder_bi = tf.keras.layers.Bidirectional(self.decoder_gru)

        self.decoder_dense = tf.keras.layers.Dense(
            self.num_classes, activation=tf.nn.softmax
        )

    def call(self, inputs):
        decoder_outputs = inputs["decoder_outputs"]
        state_f = inputs["state_f"]
        state_b = inputs["state_b"]
        encoder_outputs = inputs["encoder_outputs"]

        states = [state_f, state_b]

        attention_inputs = {"query": decoder_outputs, "values": encoder_outputs}

        context_vector = self.attention(attention_inputs)

        context_vector = tf.expand_dims(context_vector, axis=1)

        x = context_vector

        (
            decoder_outputs,
            decoder_state_forward,
            decoder_state_backward,
        ) = self.decoder_bi(x, initial_state=states)

        x = self.decoder_dense(decoder_outputs)

        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "units": self.units,
                "num_classes": self.num_classes,
            }
        )
        return config



import torch


class TimeDeltaMetric:
    """Class to calculate the time delta metric."""

    def __init__(self, percentage=0.95, normalize=False):
        """
        Parameters
        ----------
        percentage : float, optional
            percentage of absolute error to consider, by default 0.95
        normalize : bool, optional
            whether to normalize the error by the range of y_true, by default False
        """
        self.percentage = percentage
        self.normalize = normalize

    def __call__(self, y_true: torch.Tensor, y_pred: torch.Tensor) -> torch.Tensor:
        return timedelta(y_true, y_pred, self.percentage, self.normalize)

    def __repr__(self) -> str:
        return (
            f"TimeDeltaMetric(percentage={self.percentage}, normalize={self.normalize})"
        )


def timedelta(
    y_true: torch.Tensor, y_pred: torch.Tensor, percentage=0.95, normalize=False
) -> torch.Tensor:
    """Find error value that is below 95th percentile of absolute error.
    Scale this by the range of the true values (max-min)

    Parameters
    ----------
    y_true : torch.Tensor
        ground truth
    y_pred : torch.Tensor
        predictions
    percentage : float, optional
        percentage of absolute error to consider, by default 0.95
    normalize : bool, optional
        whether to normalize the error by the range of y_true, by default False

    Returns
    -------
    torch.Tensor
        Percentage percentile of absolute error normalized by range of y_true, if normalize is True
    """
    mark_percentile = int(y_true.shape[0] * percentage)
    abs_error = torch.abs(y_true - y_pred)
    delta = torch.sort(abs_error)[0][mark_percentile - 1]

    if normalize:
        norm_range = torch.max(y_true) - torch.min(y_true)
        return delta / norm_range

    return delta



import importlib
import json
import logging
import os
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Union

from datasets import Dataset, DatasetDict, Sequence, Value, load_dataset

from .dataset_config import DatasetConfig
from .dataset_utils import EncodingScheme, get_num_processors, resolve_num_proc
from .processing.feature_extractors import (
    AVAILABLE_FEATURE_EXTRACTORS,
    FEATURE_EXTRACTORS_PARAMETERS,
    LookupFeatureExtractor,
)
from .processing.processors import (
    FunctionProcessor,
    PeptideDatasetBaseProcessor,
    SequenceEncodingProcessor,
    SequencePaddingProcessor,
    SequenceParsingProcessor,
    SequencePTMRemovalProcessor,
)

logger = logging.getLogger(__name__)


class PeptideDataset:
    """
    PeptideDataset class to handle peptide datasets for deep learning models.
    The class is designed to handle peptide datasets in various formats and process them into a format that can be used by deep learning models.
    The class is built on top of the Hugging Face datasets library and provides a simple interface to load, process and save peptide datasets.

    Parameters
    ----------
    data_source : Union[str, List]
        Path to the data source file or list of paths to the data source files.
    val_data_source : Union[str, List]
        Path to the validation data source file or list of paths to the validation data source files.
    test_data_source : Union[str, List]
        Path to the test data source file or list of paths to the test data source files.
    data_format : str
        Format of the data source file(s). Example formats are 'csv', 'json', 'parquet', etc. Use 'hub' for datasets from the Hugging Face Hub and 'hf' for in-memory HF Dataset/DatasetDict objects.
    sequence_column : str
        Name of the column in the data source file that contains the peptide sequences.
    label_column : Union[str, List]
        Name of the column(s) in the data source file that contains the labels.
    val_ratio : float
        Ratio of the validation data to the training data. The value should be between 0 and 1.
    max_seq_len : int
        Maximum sequence length to pad the sequences to. If set to 0, the sequences will not be padded.
    dataset_type : str
        Type of the tensor dataset to be generated afterwards. Possible values are "tf" and "pt" for TensorFlow and PyTorch, respectively. Fallback is to TensorFlow dataset tensors.
    batch_size : int
        Batch size for the tensor dataset.
    model_features : List[str]
        List of column names in the data source file that contain features to be used by the model.
    dataset_columns_to_keep : Optional[List[str]]
        List of column names in the data source file that should be kept in the Hugging Face dataset but not returned as tensors.
    features_to_extract : Optional[List[Union[Callable, str]]]
        List of feature extractors to be applied to the sequences. The feature extractors can be either a function or a string that corresponds to a predefined feature extractor.
    pad : bool
        Flag to indicate whether to pad the sequences to the maximum sequence length.
    padding_value : str
        Value to use for padding the sequences.
    alphabet : Dict
        Alphabet to use for encoding the amino acids in the sequences.
    with_termini : bool
        Flag to indicate whether to include the N- and C-termini []- and -[] in the sequences.
    encoding_scheme : Union[str, EncodingScheme]
        Encoding scheme to use for encoding the sequences. Possible values are "unmod" and "naive-mods" for unmodified sequences and sequences with PTMs respectively.
    processed : bool
        Flag to indicate whether the dataset has been processed or not.
    enable_tf_dataset_cache : bool
        Flag to indicate whether to enable TensorFlow Dataset caching (call `.cache()` on the generated TF Datasets).
    disable_cache : bool
        Flag to indicate whether to disable Hugging Face Datasets caching. Default is False.
    auto_cleanup_cache : bool
        Flag to indicate whether to automatically clean up the temporary Hugging Face Datasets cache files. Default is True.
    num_proc : Optional[int]
        Number of processes to use for processing the dataset.
        Set to ``-1`` to use all available processors, ``None`` to force single-process execution,
        or a positive integer to use an explicit number of processors.
    batch_processing_size : Optional[int]
        Batch size for processing the dataset, passed to the HuggingFace `Dataset.map()` function calls. Default is 1000.

    Attributes
    ----------
    DEFAULT_SPLIT_NAMES : List[str]
        Default split names for the dataset.
    CONFIG_JSON_NAME : str
        Name of the configuration JSON file.

    Methods
    -------
    save_to_disk(path: str)
        Save the dataset to disk.
    load_from_disk(path: str)
        Load the dataset from disk.
    from_dataset_config(config: DatasetConfig)
        Create a PeptideDataset object from a DatasetConfig object.
    """

    DEFAULT_SPLIT_NAMES = ["train", "val", "test"]
    CONFIG_JSON_NAME = "dlomix_peptide_dataset_config.json"
    METADATA_JSON_NAME = "dlomix_peptide_dataset_metadata.json"
    SERIALIZATION_VERSION = "0.1"
    PADDING_VALUE_DEFAULT_INDEX = 0

    def __init__(self, dataset_config: DatasetConfig, **kwargs):
        super(PeptideDataset, self).__init__()
        self.__dict__.update(**dataset_config.__dict__)
        self._kwargs = kwargs

        # to be kept in the hf dataset, but not returned in the tensor dataset
        if dataset_config.dataset_columns_to_keep is None:
            self.dataset_columns_to_keep = []
        else:
            self.dataset_columns_to_keep = dataset_config.dataset_columns_to_keep

        self.encoding_scheme = EncodingScheme(dataset_config.encoding_scheme)

        self._set_hf_cache_management()

        self.extended_alphabet = {}
        self.learning_alphabet_mode = True

        if self.alphabet:
            self.extended_alphabet = self.alphabet.copy()
            self.learning_alphabet_mode = False

        # add padding value to the alphabet if not present
        if self.extended_alphabet.get(self.padding_value) is None:
            self.extended_alphabet[self.padding_value] = (
                PeptideDataset.PADDING_VALUE_DEFAULT_INDEX
            )

        self._config = dataset_config

        # explcit assignments of processed attribute
        self.processed = dataset_config.processed
        if not self.processed:
            self.hf_dataset: Optional[Union[Dataset, DatasetDict]] = None
            self._empty_dataset_mode = False
            self._is_predefined_split = False
            self._test_set_only = False
            self._num_proc = dataset_config.num_proc
            self._set_num_proc()

            self._data_files_available_splits = {}
            self._load_dataset()
            self._decide_on_splitting()

            self._relevant_columns = []
            self._extracted_features_columns = []

            if not self._empty_dataset_mode:
                self._processors = []
                self._remove_unnecessary_columns()
                self._split_dataset()
                self._parse_sequences()

                self._configure_processing_pipeline()
                self._apply_processing_pipeline()
                if self.model_features is not None:
                    self._cast_model_feature_types_to_float()
                self._cleanup_temp_dataset_cache_files()
                self.processed = True

    def _set_num_proc(self):
        n_processors = get_num_processors()
        self._num_proc, was_capped = resolve_num_proc(self._num_proc, n_processors)

        if was_capped:
            warnings.warn(
                f"Number of processors provided is greater than the available processors. Using the maximum number of processors available: {n_processors}."
            )

    def _set_hf_cache_management(self):
        if self.disable_cache:
            from datasets import disable_caching

            disable_caching()

    def _load_dataset(self):
        if self.data_format == "hub":
            self._load_from_hub()
            return

        if self.data_format == "hf":
            self._load_from_inmemory_hf_dataset()
            return

        data_sources = [self.data_source, self.val_data_source, self.test_data_source]

        for split_name, source in zip(PeptideDataset.DEFAULT_SPLIT_NAMES, data_sources):
            if source is not None:
                self._data_files_available_splits[split_name] = source

        if len(self._data_files_available_splits) == 0:
            self._empty_dataset_mode = True
            warnings.warn(
                "No data files provided, please provide at least one data source if you plan to use this dataset directly. Otherwise, you can later load data into this empty dataset"
            )
        else:
            self._empty_dataset_mode = False

            self.hf_dataset = load_dataset(
                self.data_format, data_files=self._data_files_available_splits
            )

    def _load_from_hub(self):
        self.hf_dataset = load_dataset(self.data_source, **self._kwargs)
        self._empty_dataset_mode = False
        self._is_predefined_split = True
        warnings.warn(
            'The provided data is assumed to be hosted on the Hugging Face Hub since data_format is set to "hub". Validation and test data sources will be ignored.'
        )
        if isinstance(self.hf_dataset, DatasetDict):
            for split in self.hf_dataset.keys():
                if split not in PeptideDataset.DEFAULT_SPLIT_NAMES:
                    raise ValueError(
                        f"The split name {split} is not a valid split name. Please use one of the default split names: {PeptideDataset.DEFAULT_SPLIT_NAMES}."
                    )
            self._data_files_available_splits = {
                split: f"HF hub dataset - {self.data_source} - {split}"
                for split in self.hf_dataset
            }

        else:
            self._data_files_available_splits = {
                PeptideDataset.DEFAULT_SPLIT_NAMES[
                    0
                ]: f"HF hub dataset - {self.data_source}"
            }

    def _load_from_inmemory_hf_dataset(self):
        self._empty_dataset_mode = False
        self._is_predefined_split = True
        warnings.warn(
            f'The provided data is assumed to be an in-memory Hugging Face Dataset or DatasetDict object since data_format is set to "hf". Validation and test data sources will be ignored and the split names of the DatasetDict has to follow the default namings {PeptideDataset.DEFAULT_SPLIT_NAMES}.'
        )

        if isinstance(self.data_source, DatasetDict):
            self.hf_dataset = self.data_source
            self._data_files_available_splits = dict.fromkeys(self.hf_dataset)
            self._data_files_available_splits = {
                split: f"in-memory Dataset object - {split}"
                for split in self.hf_dataset
            }

        elif isinstance(self.data_source, Dataset):
            self.hf_dataset = DatasetDict()
            self.hf_dataset[PeptideDataset.DEFAULT_SPLIT_NAMES[0]] = self.data_source
            self._data_files_available_splits = {
                PeptideDataset.DEFAULT_SPLIT_NAMES[0]: "in-memory Dataset object"
            }
        else:
            raise ValueError(
                "The provided data source is not a valid Hugging Face Dataset/DatasetDict object. The data_format value should be set to 'hf' if you plan to use an in-memory Hugging Face Dataset/DatasetDict object."
            )

    def _decide_on_splitting(self):
        count_loaded_data_sources = len(self._data_files_available_splits)

        # one data source provided -> if test, then test only, if val, then do not split
        if count_loaded_data_sources == 1:
            if (
                self.test_data_source is not None
                or PeptideDataset.DEFAULT_SPLIT_NAMES[2]
                in self._data_files_available_splits
            ):
                # test data source provided OR hugging face dataset with test split only
                self._test_set_only = True
            if self.val_data_source is not None:
                self._is_predefined_split = True

        # two or more data sources provided -> no splitting in all cases
        if count_loaded_data_sources >= 2:
            self._is_predefined_split = True

        if self._is_predefined_split:
            warnings.warn(f"""
                Multiple data sources or a single non-train data source provided {self._data_files_available_splits}, please ensure that the data sources are already split into train, val and test sets
                since no splitting will happen. If not, please provide only one data_source and set the val_ratio to split the data into train and val sets."
                """)

    def _remove_unnecessary_columns(self):
        self._relevant_columns = [self.sequence_column, *self.label_column]

        if self.model_features is not None:
            self._relevant_columns.extend(self.model_features)

        if self.dataset_columns_to_keep is not None:
            # additional columns to keep in the hugging face dataset only and not return as tensors
            self._relevant_columns.extend(self.dataset_columns_to_keep)

        # select only relevant columns from the Hugging Face Dataset (includes label column)
        self.hf_dataset = self.hf_dataset.select_columns(self._relevant_columns)

    def _split_dataset(self):
        if self._is_predefined_split or self._test_set_only:
            return

        # only a train dataset or a train and a test but no val -> split train into train/val

        splitted_dataset = self.hf_dataset[
            PeptideDataset.DEFAULT_SPLIT_NAMES[0]
        ].train_test_split(test_size=self.val_ratio)

        self.hf_dataset["train"] = splitted_dataset["train"]
        self.hf_dataset["val"] = splitted_dataset["test"]

        del splitted_dataset["train"]
        del splitted_dataset["test"]
        del splitted_dataset

    def _parse_sequences(self):
        # parse sequence in all encoding schemes
        sequence_parsing_processor = SequenceParsingProcessor(
            self.sequence_column,
            batched=True,
            with_termini=self.with_termini,
        )

        self.dataset_columns_to_keep.extend(
            SequenceParsingProcessor.PARSED_COL_NAMES.values()
        )
        self._processors.append(sequence_parsing_processor)

        if self.encoding_scheme == EncodingScheme.UNMOD:
            warnings.warn(
                f"""Encoding scheme is {self.encoding_scheme}, this enforces removing all occurences of PTMs in the sequences.
If you prefer to encode the (amino-acids)+PTM combinations as tokens in the vocabulary, please use the encoding scheme 'naive-mods'.
"""
            )

            self._processors.append(
                SequencePTMRemovalProcessor(
                    sequence_column_name=self.sequence_column, batched=True
                )
            )

    def _configure_processing_pipeline(self):
        self._configure_encoding_step()
        self._configure_padding_step()
        self._configure_feature_extraction_step()

    def _configure_encoding_step(self):
        encoding_processor = None

        if (
            self.encoding_scheme == EncodingScheme.UNMOD
            or self.encoding_scheme == EncodingScheme.NAIVE_MODS
        ):
            encoding_processor = SequenceEncodingProcessor(
                sequence_column_name=self.sequence_column,
                alphabet=self.extended_alphabet,
                batched=True,
                extend_alphabet=self.learning_alphabet_mode,
            )
        else:
            raise NotImplementedError(
                f"Encoding scheme {self.encoding_scheme} is not implemented. Available encoding schemes are: {list(EncodingScheme.__members__)}."
            )

        self._processors.append(encoding_processor)

    def _configure_padding_step(self):
        if not self.pad:
            warnings.warn(
                "Padding is turned off, sequences will have variable lengths. Converting this dataset to tensors will cause errors unless proper stacking of examples is done."
            )
            return

        padding_processor = SequencePaddingProcessor(
            sequence_column_name=self.sequence_column,
            batched=True,
            padding_index=self.extended_alphabet[self.padding_value],
            max_length=self.max_seq_len + 2 if self.with_termini else self.max_seq_len,
        )

        self._processors.append(padding_processor)

    def _configure_feature_extraction_step(self):
        if self.features_to_extract is None or len(self.features_to_extract) == 0:
            return

        for feature in self.features_to_extract:
            if isinstance(feature, str):
                feature_name = feature.lower()

                if feature_name not in AVAILABLE_FEATURE_EXTRACTORS:
                    warnings.warn(
                        f"Skipping feature extractor {feature} since it is not available. Please choose from the available feature extractors: {AVAILABLE_FEATURE_EXTRACTORS}."
                    )
                    continue

                # We pass here the parsed sequence to the feature extractor since it will always be a list with AA+PTM as elements
                feature_extactor = LookupFeatureExtractor(
                    sequence_column_name=SequenceParsingProcessor.PARSED_COL_NAMES[
                        "seq"
                    ],
                    feature_column_name=feature_name,
                    **FEATURE_EXTRACTORS_PARAMETERS[feature_name],
                    max_length=(
                        self.max_seq_len + 2 if self.with_termini else self.max_seq_len
                    ),
                    batched=True,
                )
            elif isinstance(feature, Callable):
                warnings.warn(
                    (
                        f"Using custom feature extractor from the user function {feature.__name__}"
                        "please ensure that the provided function pads the feature to the sequence length"
                        "so that all tensors have the same sequence length dimension."
                    )
                )

                feature_name = feature.__name__
                feature_extactor = FunctionProcessor(feature)
            else:
                raise ValueError(
                    f"Feature extractor {feature} is not a valid type. Please provide a function or a string that is a valid feature extractor name."
                )

            self._extracted_features_columns.append(feature_name)
            self._processors.append(feature_extactor)

    def _apply_processing_pipeline(self):
        for processor in self._processors:
            split_order = self._get_split_processing_order(processor)
            for split in split_order:
                logger.info(
                    "Applying step: %s on split %s...",
                    processor.__class__.__name__,
                    split,
                )

                logger.debug(
                    "Applying step with arguments:\n\n %s on split %s", processor, split
                )

                # split-specific logic for encoding
                if isinstance(processor, SequenceEncodingProcessor):
                    if split in PeptideDataset.DEFAULT_SPLIT_NAMES[0:2]:
                        # train/val split -> learn the alphabet unless otherwise specified
                        # force single processor to ensure that the alphabet is learned on the entire split and not just on a subset of the data when using multi-processing

                        strictly_single_process = bool(
                            getattr(processor, "extend_alphabet", False)
                            or getattr(processor, "learning_alphabet_mode", False)
                        )

                        self._apply_processor_to_split(
                            processor,
                            split,
                            force_single_processor=strictly_single_process,
                        )

                        self.extended_alphabet = processor.alphabet.copy()

                    elif split == PeptideDataset.DEFAULT_SPLIT_NAMES[2]:
                        # test split -> use the learned alphabet from the train/val split
                        # and enable fallback to encoding unseen (AA, PTM) as unmodified Amino acids
                        temp_test_processor = SequenceEncodingProcessor(
                            sequence_column_name=self.sequence_column,
                            alphabet=self.extended_alphabet,
                            batched=True,
                            extend_alphabet=False,
                            fallback_unmodified=True,
                        )

                        self._apply_processor_to_split(temp_test_processor, split)

                    else:
                        raise Warning(
                            f"When applying processors, found split '{split}' which is not a valid split name. Please use one of the default split names: {PeptideDataset.DEFAULT_SPLIT_NAMES} to ensure correct behavior."
                        )
                else:
                    # --------------------------------------------------------------------
                    # split-agnostic logic -> run processor for all splits
                    self._apply_processor_to_split(processor, split)
                    # --------------------------------------------------------------------

                # split-specific logic for truncating train/val sequences only after padding
                if isinstance(processor, SequencePaddingProcessor):
                    if split != PeptideDataset.DEFAULT_SPLIT_NAMES[2]:
                        logger.info("Removing truncated sequences in the %s ", split)

                        self.hf_dataset[split] = self.hf_dataset[split].filter(
                            lambda batch: batch[processor.KEEP_COLUMN_NAME],
                            batched=True,
                            num_proc=self._num_proc,
                            batch_size=self.batch_processing_size,
                        )

                logger.info("Done with step: %s \n", processor.__class__.__name__)

        self.hf_dataset = self.hf_dataset.remove_columns(
            SequencePaddingProcessor.KEEP_COLUMN_NAME
        )

    def _get_split_processing_order(self, processor: PeptideDatasetBaseProcessor):
        split_order = list(self.hf_dataset.keys())

        # Ensure encoding sees train/val before test so test uses full learned vocab.
        if isinstance(processor, SequenceEncodingProcessor):
            ordered_default = [
                split
                for split in PeptideDataset.DEFAULT_SPLIT_NAMES
                if split in self.hf_dataset
            ]
            non_default = [
                split for split in split_order if split not in ordered_default
            ]
            return ordered_default + non_default

        return split_order

    def _apply_processor_to_split(
        self,
        processor: PeptideDatasetBaseProcessor,
        split: str,
        force_single_processor: bool = False,
    ):
        self.hf_dataset[split] = self.hf_dataset[split].map(
            processor,
            desc=f"Mapping {processor.__class__.__name__}",
            batched=processor.batched,
            batch_size=self.batch_processing_size,
            num_proc=None if force_single_processor else self._num_proc,
        )

    def _cast_model_feature_types_to_float(self):
        def cast_to_float(feature):
            """Recursively casts Sequence and Value features to float32."""
            if isinstance(feature, Sequence):
                # Recursively apply the transformation to the nested feature
                return Sequence(cast_to_float(feature.feature))
            if isinstance(feature, Value):
                return Value("float32")
            return feature  # Return as is for unsupported feature types

        for split in self.hf_dataset.keys():
            new_features = self.hf_dataset[split].features.copy()

            for feature_name, feature_type in self.hf_dataset[split].features.items():
                # Ensure model features are casted to float for concatenation later
                if feature_name not in self.model_features:
                    continue
                new_features[feature_name] = cast_to_float(feature_type)

            self.hf_dataset[split] = self.hf_dataset[split].cast(
                new_features,
                num_proc=self._num_proc,
                batch_size=self.batch_processing_size,
            )

    def _cleanup_temp_dataset_cache_files(self):
        if self.auto_cleanup_cache:
            cleaned_up = self.hf_dataset.cleanup_cache_files()
            logger.info("Cleaned up cache files: %s.", cleaned_up)

    def _get_serializable_state(self) -> Dict[str, Any]:
        """
        Extract all state needed to reconstruct the object.

        Returns
        -------
        Dict[str, Any]
            Complete object state.
        """
        # Start with original config parameters
        state = {}

        # Add all runtime attributes (computed during processing)
        exclude = {"hf_dataset", "_config", "data_source", "_processors"}

        for key, value in self.__dict__.items():
            if key in exclude:
                continue
            # Only serialize JSON-compatible types
            try:
                json.dumps(value)
                state[key] = value
            except:
                # For complex objects, store just the type
                if isinstance(value, dict):
                    state[key] = value
                elif isinstance(value, (list, tuple, set)):
                    state[key] = list(value)
                else:
                    state[key] = str(type(value))

        return state

    def save_to_disk(self, path: str, overwrite: bool = False):
        """
        Save the dataset to disk.

        Parameters
        ----------
        path : str
            Path to save the dataset to.
        overwrite : bool, optional
            Whether to overwrite existing directory, by default False.

        """

        path_obj = Path(path)

        # Check if path exists
        if path_obj.exists() and not overwrite:
            raise FileExistsError(
                f"Directory {path} already exists. Set overwrite=True to overwrite and replace the saved dataset."
            )

        # Create directory
        path_obj.mkdir(parents=True, exist_ok=True)

        # 1. Save the ORIGINAL config (input parameters only)
        # No refresh needed - config represents what user provided
        self._config.save_config_json(os.path.join(path, self.CONFIG_JSON_NAME))

        # 2. Save the complete object state (metadata)
        state = self._get_serializable_state()
        metadata = {
            "class_name": self.__class__.__name__,
            "module_name": self.__class__.__module__,
            "state": state,
            "version": PeptideDataset.SERIALIZATION_VERSION,  # Version for serialization format
        }

        with open(path_obj / self.METADATA_JSON_NAME, "w") as f:
            json.dump(metadata, f, indent=2)

        # 3. Save the HuggingFace dataset
        if self.hf_dataset is not None:
            self.hf_dataset.save_to_disk(str(path_obj / "hf_dataset"))

        return True

    def _validate_loaded_state(self):
        """
        Validate that the loaded state is consistent.

        Raises
        ------
        ValueError
            If the state is inconsistent.
        """
        # Check that essential attributes exist
        if not hasattr(self, "hf_dataset") or self.hf_dataset is None:
            raise ValueError("HuggingFace dataset not loaded properly.")

        if not self.processed:
            raise ValueError("Dataset should be marked as processed after loading.")

        # Validate that dataset columns match expected columns
        if hasattr(self, "_relevant_columns"):
            for split in self.hf_dataset.keys():
                dataset_columns = set(self.hf_dataset[split].column_names)
                expected_columns = set(self._relevant_columns)

                if not expected_columns.issubset(dataset_columns):
                    missing = expected_columns - dataset_columns
                    raise ValueError(
                        f"Split '{split}' is missing expected columns: {missing}"
                    )

    @classmethod
    def from_dataset_config(cls, config: DatasetConfig):
        return cls(**config.__dict__)

    def __getitem__(self, index):
        return self.hf_dataset[index]

    def __str__(self):
        return self.hf_dataset.__str__()

    def __repr__(self):
        return self.hf_dataset.__repr__()

    def __len__(self):
        return len(self.hf_dataset)

    def __iter__(self):
        return iter(self.hf_dataset)

    def __getattr__(self, attr):
        if attr not in self.__dict__:
            return getattr(self.hf_dataset, attr)
        else:
            return self.__dict__[attr]

    def _get_input_tensor_column_names(self):
        # return a list of columns to be used as input tensors
        input_tensor_columns = self._relevant_columns.copy()

        # remove the label column(s) from the input tensor columns since the to_tf_dataset method has a separate label_cols argument
        for label in self.label_column:
            input_tensor_columns.remove(label)

        # remove the columns that are not needed in the tensor dataset
        input_tensor_columns = list(
            set(input_tensor_columns) - set(self.dataset_columns_to_keep)
        )

        # add the extracted features columns to the input tensor columns
        input_tensor_columns.extend(self._extracted_features_columns)

        return input_tensor_columns

    @property
    def tensor_train_data(self):
        """TensorFlow or Torch Dataset object for the training data"""
        if self.dataset_type == "pt":
            return self._get_split_torch_dataset(PeptideDataset.DEFAULT_SPLIT_NAMES[0])
        else:
            tf_dataset = self._get_split_tf_dataset(
                PeptideDataset.DEFAULT_SPLIT_NAMES[0]
            )

            if self.enable_tf_dataset_cache:
                tf_dataset = tf_dataset.cache()

            return tf_dataset

    @property
    def tensor_val_data(self):
        """TensorFlow or Torch Dataset object for the val data"""
        if self.dataset_type == "pt":
            return self._get_split_torch_dataset(PeptideDataset.DEFAULT_SPLIT_NAMES[1])
        else:
            tf_dataset = self._get_split_tf_dataset(
                PeptideDataset.DEFAULT_SPLIT_NAMES[1]
            )

            if self.enable_tf_dataset_cache:
                tf_dataset = tf_dataset.cache()

            return tf_dataset

    @property
    def tensor_test_data(self):
        """TensorFlow or Torch Dataset object for the test data"""
        if self.dataset_type == "pt":
            return self._get_split_torch_dataset(PeptideDataset.DEFAULT_SPLIT_NAMES[2])
        else:
            tf_dataset = self._get_split_tf_dataset(
                PeptideDataset.DEFAULT_SPLIT_NAMES[2]
            )

            return tf_dataset

    def _check_if_split_exists(self, split_name: str):
        existing_splits = list(self.hf_dataset.keys())
        if split_name not in existing_splits:
            raise ValueError(
                f"Split '{split_name}' does not exist in the dataset. Available splits are: {existing_splits}"
            )
        return True

    def _get_split_tf_dataset(self, split_name: str):
        self._check_if_split_exists(split_name)

        # to return a tuple if it is a single label column and be compatible with HF Datasets API updates

        label_cols = self.label_column

        if isinstance(self.label_column, list) and len(self.label_column) == 1:
            label_cols = self.label_column[0]

        return self.hf_dataset[split_name].to_tf_dataset(
            columns=self._get_input_tensor_column_names(),
            label_cols=label_cols,
            shuffle=(
                self.shuffle
                if split_name == PeptideDataset.DEFAULT_SPLIT_NAMES[0]
                else False
            ),
            batch_size=self.batch_size,
        )

    def _get_split_torch_dataset(self, split_name: str):
        self._check_if_split_exists(split_name)

        from torch.utils.data import DataLoader

        # Prepare DataLoader kwargs, starting with defaults
        dataloader_kwargs = {
            "dataset": self.hf_dataset[split_name].with_format(
                type="torch",
                columns=[*self._get_input_tensor_column_names(), *self.label_column],
            ),
            "batch_size": self.batch_size,
            "shuffle": (
                self.shuffle
                if split_name == PeptideDataset.DEFAULT_SPLIT_NAMES[0]
                else False
            ),
        }

        # Update with user-provided torch_dataloader_kwargs if available
        if hasattr(self, "torch_dataloader_kwargs") and self.torch_dataloader_kwargs:
            # Don't override the dataset parameter
            user_kwargs = {
                k: v for k, v in self.torch_dataloader_kwargs.items() if k != "dataset"
            }
            dataloader_kwargs.update(user_kwargs)

        data_loader = DataLoader(**dataloader_kwargs)

        return data_loader


# ------------------------------------------------------------------------------------------------


def load_processed_dataset(path: str, validate: bool = True):
    """
    Load a processed peptide dataset from a given path.

    Parameters
    ----------
    path : str
        Path to the peptide dataset.
    validate : bool, optional
        Whether to validate the loaded dataset state, by default True.

    Returns
    -------
    dlomix.data.PeptideDataset or one of its child classes
        Peptide dataset.
    """

    path_obj = Path(path)

    if not path_obj.exists():
        raise FileNotFoundError(
            f"Provided directory for loading the dataset: {path} does not exist."
        )

    # 1. Load the config
    config = DatasetConfig.load_config_json(
        str(path_obj / PeptideDataset.CONFIG_JSON_NAME)
    )

    # 2. Load metadata
    metadata_path = path_obj / PeptideDataset.METADATA_JSON_NAME
    if metadata_path.exists():
        with open(metadata_path, "r") as f:
            metadata = json.load(f)
    else:
        # Fallback for datasets saved without metadata
        metadata = None

    # 3. Create instance with processed=True to skip initialization processing
    module = importlib.import_module("dlomix.data")
    class_name = metadata.get("class_name") if metadata else "PeptideDataset"
    cls = getattr(module, class_name)

    # ensure processed flag is set in config to avoid re-processing
    config.processed = True

    instance = cls.from_dataset_config(config)
    instance._config.processed = False  # Reset config processed flag
    instance.processed = True  # Ensure processed flag is set

    # 4. Restore state from metadata
    if metadata:
        for key, value in metadata["state"].items():
            setattr(instance, key, value)

    # 5. Load the HuggingFace dataset
    hf_dataset_path = path_obj / "hf_dataset"
    if hf_dataset_path.exists():
        from datasets import load_from_disk

        instance.hf_dataset = load_from_disk(str(hf_dataset_path))

    # 6. Validate if requested
    if validate:
        instance._validate_loaded_state()

    return instance



from ..config import _BACKEND, PYTORCH_BACKEND, TENSORFLOW_BACKEND

__all__ = []

if _BACKEND in TENSORFLOW_BACKEND:
    from .attention import AttentionLayer, DecoderAttentionLayer
elif _BACKEND in PYTORCH_BACKEND:
    from .attention_torch import AttentionLayer, DecoderAttentionLayer
    from .bi_gru_seq_encoder_torch import BiGRUSequentialEncoder

    __all__.append("BiGRUSequentialEncoder")

__all__.extend(
    [
        "AttentionLayer",
        "DecoderAttentionLayer",
    ]
)



import os
import re

import numpy as np
import pandas as pd
import wandb
import wandb.apis.reports as wr

from ..data.retention_time import RetentionTimeDataset


class RetentionTimeReportModelComparisonWandb:
    """Creates a WandB report for comparing models.

    Parameters
    ----------
    models : dict
        A dictionary where the keys are model names and the values are model objects.
    project : str
        The name of the project.
    title : str
        The title of the report.
    description : str
        The description of the report.
    test_dataset : RetentionTimeDataset
        The test dataset object to compare the predictions of models on.
    """

    # Wilhelmlab WandB account that has all VEGA presets required for the reports
    VEGA_LITE_PRESETS_ID = "prosit-compms"

    def __init__(
        self,
        models: dict,
        project: str,
        title: str,
        description: str,
        test_dataset: RetentionTimeDataset,
    ):
        self.project = project
        self.title = title
        self.description = description
        self.models = models
        self.test_dataset = test_dataset
        self.entity = wandb.apis.PublicApi().default_entity
        self.api = wandb.Api()

    def create_report(
        self,
        add_data_section=True,
        add_residuals_section=True,
        add_r2_section=True,
        add_density_section=True,
    ):
        """Creates the report in wandb.

        Parameters
        ----------
            add_data_section: bool, optional
                Add a section for input data to the report. Defaults to True.
            add_residuals_section: bool, optional
                Add a section for residual plots. Defaults to True.
            add_r2_section: bool, optional
                Add a section for the R2 metric. Defaults to True.
            add_density_section: bool, optional
                Add a section for the density plot. Defaults to True.
        """
        report = wr.Report(
            project=self.project, title=self.title, description=self.description
        )

        report.blocks = [wr.TableOfContents()]

        if add_data_section:
            report.blocks += self._build_data_section()
        if add_residuals_section:
            report.blocks += self._build_residuals_section()
        if add_r2_section:
            report.blocks += self._build_r2_section()
        if add_density_section:
            report.blocks += self._build_density_section()

        report.save()

    def calculate_r2(self, targets, predictions):
        from sklearn.metrics import r2_score

        r2 = r2_score(targets, predictions)
        return r2

    def calculate_residuals(self, targets, predictions):
        residuals = predictions - targets
        return residuals

    def _build_data_section(self):
        data_block = [
            wr.H1(text="Data"),
            wr.P(
                "The following section is showing a simple explorative data analysis of the used dataset. The first histogram shows the distribution of peptide lengths in the data set, while the second histogram shows the distribution of indexed retention times."
            ),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=[
                    wr.CustomChart(
                        query={"summaryTable": {"tableKey": self.table_key_len}},
                        chart_name=f"{RetentionTimeReportModelComparisonWandb.VEGA_LITE_PRESETS_ID}/histogram_peptide_length",
                        chart_fields={"value": self.test_dataset.sequence_col},
                    ),
                    wr.CustomChart(
                        query={"summaryTable": {"tableKey": self.table_key_rt}},
                        chart_name=f"{RetentionTimeReportModelComparisonWandb.VEGA_LITE_PRESETS_ID}/histogram_irt",
                        chart_fields={"value": self.test_dataset.target_col},
                    ),
                ],
            ),
            wr.HorizontalRule(),
        ]
        return data_block

    def _build_residuals_section(self):
        panel_list_models = []
        for model in self.models:
            panel_list_models.append(
                wr.CustomChart(
                    query={"summaryTable": {"tableKey": f"results_table_{model}"}},
                    chart_name=f"{RetentionTimeReportModelComparisonWandb.VEGA_LITE_PRESETS_ID}/histogram_residuals_irt",
                    chart_fields={"value": "residuals", "name": model},
                )
            )

        residuals_block = [
            wr.H1(text="Residuals"),
            wr.P(
                "This section shows the residuals histograms. Each plot shows the residuals of each of the compared models"
            ),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=panel_list_models,
            ),
            wr.HorizontalRule(),
        ]

        return residuals_block

    def _build_r2_section(self):
        r2_block = [
            wr.H1(text="R2"),
            wr.P(
                "The following plot displays the R2 score for all the compared models."
            ),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=[
                    wr.BarPlot(
                        title="R2",
                        metrics=["r2"],
                        orientation="h",
                        title_x="R2",
                        max_runs_to_show=20,
                        max_bars_to_show=20,
                        font_size="auto",
                    ),
                ],
            ),
            wr.HorizontalRule(),
        ]
        return r2_block

    def _build_density_section(self, irt_delta95=5):
        panel_list_models = []
        targets = self.test_dataset.get_split_targets(
            split=self.test_dataset.main_split
        )
        for model in self.models:
            panel_list_models.append(
                wr.CustomChart(
                    query={"summaryTable": {"tableKey": f"results_table_{model}"}},
                    chart_name=f"{RetentionTimeReportModelComparisonWandb.VEGA_LITE_PRESETS_ID}/density_plot_irt",
                    chart_fields={
                        "measured": "irt",
                        "predicted": "predicted_irt",
                        "name": model,
                        "irt_delta95": irt_delta95,
                    },
                )
            )

        density_block = [
            wr.H1(text="Density"),
            wr.P("This section displays the density plots for all compared models."),
            wr.PanelGrid(
                runsets=[
                    wr.Runset(self.entity, self.project),
                ],
                panels=panel_list_models,
            ),
            wr.HorizontalRule(),
        ]

        return density_block

    def compare_models(self):
        for model in self.models:
            # initialize WANDB
            current_model = model
            wandb.init(project=self.project, name=current_model)

            # predict on test_dataset
            predictions = self.models[model].predict(self.test_dataset.test_data)
            predictions = predictions.ravel()
            targets = self.test_dataset.get_split_targets(
                split=self.test_dataset.main_split
            )
            # create result df
            results_df = pd.DataFrame(
                {
                    "sequence": self.test_dataset.sequences,
                    "irt": targets,
                    "predicted_irt": predictions,
                    "residuals": self.calculate_residuals(targets, predictions),
                }
            )
            # log df as table to wandb
            table = wandb.Table(dataframe=results_df)
            wandb.log({f"results_table_{current_model}": table})

            # log r2 to wandb
            r2 = self.calculate_r2(targets, predictions)
            wandb.log({"r2": r2})

            # finish run
            wandb.finish()

    # function to log sequence length table to wandb
    def log_sequence_length_table(
        self, data: pd.DataFrame, seq_col: str = "modified_sequence"
    ):
        name_hist = "counts_hist"
        counts = self.count_seq_length(data, seq_col)
        # convert to df for easier handling
        counts_df = counts.to_frame()
        table = wandb.Table(dataframe=counts_df)
        # log to wandb
        hist = wandb.plot_table(
            vega_spec_name=f"{RetentionTimeReportModelComparisonWandb.VEGA_LITE_PRESETS_ID}/histogram_peptide_length",
            data_table=table,
            fields={"value": seq_col},
        )
        wandb.log({name_hist: hist})
        name_hist_table = name_hist + "_table"
        return name_hist_table

    # function to count sequence length
    def count_seq_length(self, data: pd.DataFrame, seq_col: str) -> pd.Series:
        pattern = re.compile(r"\[UNIMOD:.*\]", re.IGNORECASE)
        data[seq_col].replace(pattern, "", inplace=True)
        return data[seq_col].str.len()

    # function to log retention time table to wandb
    def log_rt_table(self, data: pd.DataFrame, rt_col: str = "indexed_retention_time"):
        name_hist = "rt_hist"
        rt = data.loc[:, rt_col]
        # convert to df for easier handling
        rt_df = rt.to_frame()
        table = wandb.Table(dataframe=rt_df)
        # log to wandb
        hist = wandb.plot_table(
            vega_spec_name=f"{RetentionTimeReportModelComparisonWandb.VEGA_LITE_PRESETS_ID}/histogram_irt",
            data_table=table,
            fields={"value": rt_col},
        )
        wandb.log({name_hist: hist})
        name_hist_table = name_hist + "_table"
        return name_hist_table

    def log_data(self):
        wandb.init(project=self.project, name="data_run")
        # check if datasource is a string
        if isinstance(self.test_dataset.data_source, str):
            # read corresponding file
            file_extension = os.path.splitext(self.test_dataset.data_source)[-1].lower()
            data = pd.DataFrame()

            if file_extension == ".csv":
                data = pd.read_csv(self.test_dataset.data_source)
            if file_extension == ".json":
                data = pd.read_json(self.test_dataset.data_source)
            if file_extension == ".parquet":
                data = pd.read_parquet(
                    self.test_dataset.data_source, engine="fastparquet"
                )

            self.table_key_len = self.log_sequence_length_table(
                data, self.test_dataset.sequence_col
            )
            self.table_key_rt = self.log_rt_table(data, self.test_dataset.target_col)

        # check if datasource is a tuple of two ndarrays or two lists
        if (
            isinstance(self.test_dataset.data_source, tuple)
            and all(
                isinstance(item, (np.ndarray, list))
                for item in self.test_dataset.data_source
            )
            and len(self.test_dataset.data_source) == 2
        ):
            data = pd.DataFrame(
                {
                    self.test_dataset.sequence_col: self.test_dataset.data_source[0],
                    self.test_dataset.target_col: self.test_dataset.data_source[1],
                }
            )
            self.table_key_len = self.log_sequence_length_table(
                data, self.test_dataset.sequence_col
            )
            self.table_key_rt = self.log_rt_table(data, self.test_dataset.target_col)

        # check if datasource is a single ndarray or list
        # does not work? maybe error in RetentionTimeDataset
        if isinstance(self.test_dataset.data_source, (np.ndarray, list)):
            data = pd.DataFrame(
                {self.test_dataset.sequence_col: self.test_dataset.data_source}
            )
            self.table_key_len = self.log_sequence_length_table(
                data, self.test_dataset.sequence_col
            )
        wandb.finish()



import os
import warnings

DEFAULT_BACKEND = "tensorflow"
BACKEND_PRETTY_NAME = "TensorFlow"

TENSORFLOW_BACKEND = ["tensorflow", "tf"]
PYTORCH_BACKEND = ["pytorch", "torch", "pt"]


def custom_show_warning(msg, category, filename, lineno, file=None, line=None):
    print(f"{msg}")


# Allow setting backend via environment variable or default to tensorflow
_BACKEND = os.environ.get("DLOMIX_BACKEND", DEFAULT_BACKEND).lower().strip()

if _BACKEND in TENSORFLOW_BACKEND:
    BACKEND_PRETTY_NAME = "TensorFlow"
elif _BACKEND in PYTORCH_BACKEND:
    BACKEND_PRETTY_NAME = "PyTorch"
else:
    # default to TensorFlow
    with warnings.catch_warnings():
        warnings.simplefilter("once", category=UserWarning)  # Show only once
        warnings.showwarning = (
            custom_show_warning  # Temporarily override the default showwarning
        )
        warnings.warn(
            f"Backend '{_BACKEND}' is not supported. Defaulting to {BACKEND_PRETTY_NAME} backend.",
            UserWarning,
        )

    _BACKEND = DEFAULT_BACKEND


with warnings.catch_warnings():
    warnings.simplefilter("once", category=UserWarning)  # Show only once
    warnings.showwarning = (
        custom_show_warning  # Temporarily override the default showwarning
    )

    message = f"Using {BACKEND_PRETTY_NAME} Backend for DLOmix. To change the backend, set the DLOMIX_BACKEND environment variable to tensorflow or pytorch and re-import DLOmix."
    warnings.warn(message, UserWarning)  # Trigger the warning



import torch.nn as nn

from .attention_torch import DecoderAttentionLayer


class GRUSequentialDecoder(nn.Module):
    """Encoder class needed to handle two GRU outputs in torch.

    No implementation using nn.Sequential() possible.

    Args:
        embedding_output_dim (int): The size of the embedding output dimension. Defaults to 16.
        recurrent_layers_sizes (tuple): The sizes of the recurrent layers. Defaults to (256, 512).
        dropout_rate (float): The dropout rate used in the encoder layers. Defaults to 0.5.
    """

    def __init__(
        self,
        recurrent_layers_sizes,
        dropout_rate,
        max_ion,
    ):
        super(GRUSequentialDecoder, self).__init__()
        self.unidirectional_GRU = nn.GRU(
            input_size=recurrent_layers_sizes[1],
            hidden_size=recurrent_layers_sizes[1],
            batch_first=True,
            bidirectional=False,
        )
        self.encoder_dropout = nn.Dropout(dropout_rate)

        self.attention = DecoderAttentionLayer(max_ion)

    def forward(self, inputs):
        x, _ = self.unidirectional_GRU(inputs)
        x = self.encoder_dropout(x)
        x = self.attention(x)
        return x



import logging

import pytest

from dlomix.models.chargestate import ChargeStatePredictor
from dlomix.models.prosit import PrositIntensityPredictor, PrositRetentionTimePredictor

logger = logging.getLogger(__name__)


def test_prosit_retention_time_model():
    model = PrositRetentionTimePredictor()
    logger.info(model)
    assert model is not None


def test_prosit_intensity_model():
    model = PrositIntensityPredictor(
        input_keys={
            "SEQUENCE_KEY": "sequence",
        },
        meta_data_keys=["collision_energy", "precursor_charge"],
    )

    seq_len = model.seq_length
    model.build(
        {
            "sequence": (
                None,
                seq_len,
            ),
            "collision_energy": (None, 1),
            "precursor_charge": (None, 6),
        }
    )
    model.summary(print_fn=logger.info)
    logger.info(model)
    assert model is not None


def test_prosit_intensity_model_ptm_on_input():
    model = PrositIntensityPredictor(
        input_keys={
            "SEQUENCE_KEY": "sequence",
        },
        meta_data_keys=["collision_energy", "precursor_charge", "fragmentation_type"],
        use_prosit_ptm_features=True,
    )

    seq_len = model.seq_length
    model.build(
        {
            "sequence": (
                None,
                seq_len,
            ),
            "collision_energy": (None, 1),
            "precursor_charge": (None, 6),
            "fragmentation_type": (None, 1),
            PrositIntensityPredictor.PTM_INPUT_KEYS[0]: (
                None,
                seq_len,
                6,
            ),
            PrositIntensityPredictor.PTM_INPUT_KEYS[1]: (
                None,
                seq_len,
                6,
            ),
            PrositIntensityPredictor.PTM_INPUT_KEYS[2]: (
                None,
                seq_len,
                1,
            ),
        }
    )
    model.summary(print_fn=logger.info)
    logger.info(model.input_keys)
    logger.info(model)
    assert model is not None


def test_prosit_intensity_model_ptm_on_missing():
    model = PrositIntensityPredictor(use_prosit_ptm_features=True)
    seq_len = model.seq_length
    with pytest.raises(ValueError, match="PTM"):
        model.build(
            {
                "sequence": (
                    None,
                    seq_len,
                ),
                "collision_energy": (None,),
                "precursor_charge": (None, 6),
                "fragmentation_type": (None,),
                # no PTM features
            }
        )


def test_prosit_intensity_model_encoding_metadata_missing():
    model = PrositIntensityPredictor(
        meta_data_keys=["meta_data_1", "meta_data_2"], use_meta_data=True
    )
    seq_len = model.seq_length

    with pytest.raises(ValueError, match="metadata"):
        model.build(
            {
                "sequence": (
                    None,
                    seq_len,
                ),
                # no meta-data while expected
            }
        )


def test_prosit_intensity_model_no_metadata():
    model = PrositIntensityPredictor(
        input_keys={
            "SEQUENCE_KEY": "sequence",
        },
        meta_data_keys=None,
    )

    seq_len = model.seq_length

    assert model.meta_data_keys == []

    model.build(
        {
            "sequence": (
                None,
                seq_len,
            ),
        }
    )

    assert model is not None
    assert model.meta_encoder is None


def basic_model_existence_test(model):
    logger.info(model)
    assert model is not None

    # Explicitly build the model with a dummy input shape (batch_size, seq_length)
    model.build((None, 30))
    assert len(model.trainable_variables) > 0


def test_dominant_chargestate_model():
    model = ChargeStatePredictor(model_flavour="dominant")
    basic_model_existence_test(model)


def test_observed_chargestate_model():
    model = ChargeStatePredictor(model_flavour="observed")
    basic_model_existence_test(model)


def test_chargestate_distribution_model():
    model = ChargeStatePredictor(model_flavour="relative")
    basic_model_existence_test(model)



import torch

EPSILON = torch.finfo(torch.float32).eps


def adjusted_mean_absolute_error(
    y_pred: torch.tensor, y_true: torch.tensor
) -> torch.tensor:
    """
    For two vectors, discard those components that
    are 0 in both vectors and compute the mean
    absolute error for the adjusted vector.
    """
    # Convert y_true and y_pred to float tensors
    y_true = y_true.to(torch.float32)
    y_pred = y_pred.to(torch.float32)

    # Create a mask for elements that are not both zero
    # for torch you can use numpy logic!
    mask = (y_true != 0) & (y_pred != 0)

    # Compute the mean absolute error
    absolute_errors = (y_true - y_pred).abs() * mask

    count_non_zero = mask.sum()

    # Avoid division by zero by adding a small epsilon to the denominator
    mean_absolute_error = (absolute_errors).sum() / (count_non_zero + EPSILON)

    return mean_absolute_error


def adjusted_mean_squared_error(
    y_pred: torch.tensor, y_true: torch.tensor
) -> torch.tensor:
    """
    For two vectors, discard those components that
    are 0 in both vectors and compute the mean
    absolute error for the adjusted vector.
    """
    # Convert y_true and y_pred to float tensors
    y_true = y_true.to(torch.float32)
    y_pred = y_pred.to(torch.float32)

    # Create a mask for elements that are not both zero
    # for torch you can use numpy logic!
    mask = (y_true != 0) & (y_pred != 0)

    # Compute the mean absolute error
    squared_errors = torch.square((y_true - y_pred) * mask)

    count_non_zero = mask.sum()

    # Avoid division by zero by adding a small epsilon to the denominator
    mean_squared_error = (squared_errors).sum() / (count_non_zero + EPSILON)

    return mean_squared_error
